import unittest
from unittest.mock import MagicMock, patch

from pyspark.sql import Row

from spetlr.deltaspec.DeltaTableSpec import DeltaTableSpec


class TestDeltaTableSpec(unittest.TestCase):
    """
    The detail schema can be found here:
    https://learn.microsoft.com/en-us/azure/databricks/delta/table-details

    """

    @patch("spetlr.spark.Spark.get")  # Mock Spark session retrieval
    def test_from_name_with_mocked_describe_detail(self, mock_spark_get):
        """Test DeltaTableSpec.from_name with a mocked DESCRIBE DETAIL output."""

        # Mock Spark session
        mock_spark = MagicMock()
        mock_spark_get.return_value = mock_spark

        # Define the row data as per the example
        mock_row = Row(
            format="delta",
            id="04e1db34-399c-42d8-8c56-a7311f60e73d",
            name="test_catalog.test_db.test_table",
            description=None,
            location="abfss://catalog@heyhey.dfs.core.windows.net/__unitystorage/"
            "catalogs/xxx/tables/yyyy",
            createdAt=None,  # Set as None since timestamps are not needed in this test
            lastModified=None,
            partitionColumns=[],
            clusteringColumns=[],
            numFiles=0,
            sizeInBytes=0,
            properties={"delta.enableDeletionVectors": "true"},
            minReaderVersion=3,
            minWriterVersion=7,
            tableFeatures=["deletionVectors"],
            statistics={"numRowsDeletedByDeletionVectors": 0, "numDeletionVectors": 0},
        )

        # Ensure that
        # spark.sql("DESCRIBE DETAIL ...").collect()
        # returns a **list with a single Row**
        mock_spark.sql.return_value.collect.return_value = [mock_row]

        # Run the test
        spec = DeltaTableSpec.from_name("testcatalog.test_db.test_table")

        # Assert that the location is set to None for managed tables
        self.assertIsNone(
            spec.location,
            "The location should be None " "for managed Unity Catalog tables.",
        )
        self.assertEqual(spec.name, "test_catalog.test_db.test_table")
        self.assertIsNone(spec.comment)  # Description is mapped to `comment`
        self.assertEqual(spec.partitioned_by, [])
        self.assertEqual(spec.cluster_by, [])
        self.assertEqual(
            spec.tblproperties,
            {
                "delta.enableDeletionVectors": "true",
                "delta.minReaderVersion": "3",
                "delta.minWriterVersion": "7",
            },
        )



from typing import List

import pyspark.sql.functions as f
from pyspark.sql import DataFrame


def CheckDfMerge(
    *,
    df: DataFrame,
    df_target: DataFrame,
    join_cols: List[str],
    avoid_cols: List[str],
):
    """This logic optimizes data load of dataframe, df, into a target table, df_target.
    it checks whether a merge is needed.
    If it is not needed, a simple insert is executed.
    """

    key_col = join_cols[-1]

    # Write the filtering string including casting maps to strings
    filter_string = f"b.{key_col} IS NULL"
    for col, col_typ in df.dtypes:
        if col not in join_cols + avoid_cols:
            if col_typ == "map":
                filter_string += (
                    f" OR (CAST(a.{col} AS varchar) <> CAST(b.{col} AS varchar))"
                )
            else:
                filter_string += f" OR (a.{col} <> b.{col})"
                filter_string += f" OR (a.{col} IS NULL AND b.{col} IS NOT NULL)"
                filter_string += f" OR (a.{col} IS NOT NULL AND b.{col} IS NULL)"

    # The following compares the df and df_target to check if any merging is required.
    #
    df = (
        df.alias("a")
        .join(
            df_target.alias("b").select(
                "b.*", f.col(f"b.{key_col}").alias(f"{key_col}_copy")
            ),
            on=join_cols,
            how="left",
        )
        .filter(filter_string)
        .withColumn(
            "is_new",
            f.when(f.col(f"{key_col}_copy").isNull(), True).otherwise(False),
        )
        .select("a.*", "is_new")
        .cache()
    )

    # merge_required == True if the data contains updates or deletes,
    # False if only inserts
    # note: inserts alone happen almost always and can use "append",
    # which is much faster than merge
    merge_required = len(df.filter(~f.col("is_new")).take(1)) > 0
    df = df.drop("is_new")

    return df, merge_required



from pyspark.sql import DataFrame
from pyspark.sql import functions as f

from spetlr.delta import DeltaHandle
from spetlr.etl import Transformer


class EhJsonToDeltaTransformer(Transformer):
    def __init__(self, *, target_dh: DeltaHandle, case_sensitive: bool = True):
        super().__init__()
        self.target_dh = target_dh
        self.case_sensitive = case_sensitive

    def process(self, df: DataFrame) -> DataFrame:
        # use the schema from the target table to decide what to unpack
        target_df = self.target_dh.read()
        source_df = df
        _keep_body_as_json = False

        if "BodyJson" in target_df.columns:
            _keep_body_as_json = True

        # these columns will be copied directly from the source data frame
        # but BodyJson should NOT be unpacked as a struct
        # the BodyJson is therefore removed from direct_cols
        direct_cols = [
            col
            for col in target_df.columns
            if col in source_df.columns and col != "BodyJson"
        ]
        # verify that the schema of direct columns matches
        for col in direct_cols:
            target_type = target_df.select(col).schema.fields[0].dataType
            source_type = source_df.select(col).schema.fields[0].dataType
            if target_type != source_type:
                raise TypeError(
                    "The target table has incorrect type for "
                    f"direct column {col}, "
                    f"expected {source_type}, got {target_type}."
                )

        # The body is saved as string format as "BodyJson"
        if _keep_body_as_json:
            df = df.withColumn("BodyJson", f.col("Body").cast("string"))

        if "Body" in direct_cols:
            if _keep_body_as_json:
                df = df.select(*direct_cols, "BodyJson")
            else:
                df = df.select(*direct_cols)
        else:
            json_options = {}
            if not self.case_sensitive:
                json_options["readerCaseSensitive"] = "false"

            # every column that is in the target delta table and that is not a direct
            # column from the source eventhub DataFrame, is assumed to be a column whose
            # value can be unpacked from the json that is in the eventhub body
            body_cols = [
                col
                for col in target_df.columns
                if col not in source_df.columns and col != "BodyJson"
            ]
            body_schema = target_df.select(*body_cols).schema
            df = df.withColumn(
                "Body",
                f.from_json(
                    f.decode("Body", "utf-8"), body_schema, options=json_options
                ).alias("Body"),
            )
            if _keep_body_as_json:
                df = df.select("Body.*", *direct_cols, "BodyJson")
            else:
                df = df.select("Body.*", *direct_cols)

        return df



# Defines a class for opening a connection to Cosmos DB,
# loading and saving tables, and executing SQL.
import hashlib

# e.g. usage:
#   server = CosmosDb()
#   table_name = CosmosDb.table_name("TableId")
#   server.execute_sql(
#       f"CREATE TABLE cosmosCatalog.database_name.table_name using cosmos.oltp"
#       )
#   df = server.load_table("TableId")
#   server.save_table(df, "TableId")
from typing import Optional, Union

from azure.cosmos import CosmosClient, DatabaseProxy, PartitionKey
from pyspark.sql import DataFrame
from pyspark.sql.types import DataType

from spetlr.configurator.configurator import Configurator
from spetlr.cosmos.cosmos_base_server import CosmosBaseServer
from spetlr.cosmos.cosmos_handle import CosmosHandle
from spetlr.exceptions import NoSuchSchemaException, SpetlrException
from spetlr.schema_manager import SchemaManager
from spetlr.spark import Spark


class SpetlrCosmosException(SpetlrException):
    pass


class CosmosDb(CosmosBaseServer):
    def __init__(
        self,
        account_key: str,
        database: str,
        account_name: str = None,
        endpoint: str = None,
        catalog_name: Optional[str] = "cosmosCatalog",
    ):
        if not account_name and not endpoint:
            raise ValueError("account_name or endpoint must be set")

        if not endpoint:
            endpoint_pattern = "https://{}.documents.azure.com:443/"
            endpoint = endpoint_pattern.format(account_name)

        self.endpoint = endpoint
        self.account_key = account_key
        self.database = database
        self.config = {
            "spark.cosmos.accountEndpoint": self.endpoint,
            "spark.cosmos.accountkey": account_key,
            "spark.cosmos.database": database,
            "spark.cosmos.container": None,
        }
        self.client = CosmosClient(endpoint, credential=account_key)

        self._db_client: Optional[DatabaseProxy] = None
        self.catalog_name = (
            catalog_name
            or hashlib.sha1(f"{endpoint}{self.database}".encode()).hexdigest()
        )

    @property
    def db_client(self):
        if self._db_client is not None:
            return self._db_client

        self._db_client = self.client.get_database_client(self.database)
        return self._db_client

    def execute_sql(self, sql: str):
        # TODO: Deprecate this when UC is enabled
        """
        Note that this method is not compatible with UC-enabled clusters.
        Examples:
        sql = f"CREATE DATABASE IF NOT EXISTS cosmosCatalog.{database_name};"
        sql = f"CREATE TABLE IF NOT EXISTS cosmosCatalog.{database_name}.{table_name}"
            " using cosmos.oltp "
            "TBLPROPERTIES(partitionKeyPath = '/id', manualThroughput = '1100')"
        """
        spark = Spark.get()
        spark.conf.set(
            f"spark.sql.catalog.{self.catalog_name}",
            "com.azure.cosmos.spark.CosmosCatalog",
        )
        spark.conf.set(
            f"spark.sql.catalog.{self.catalog_name}.spark.cosmos.accountEndpoint",
            self.endpoint,
        )
        spark.conf.set(
            f"spark.sql.catalog.{self.catalog_name}.spark.cosmos.accountKey",
            self.account_key,
        )
        return spark.sql(sql)

    def read_table_by_name(self, table_name: str, schema: DataType = None) -> DataFrame:
        config = self.config.copy()
        config["spark.cosmos.container"] = table_name
        rd = Spark.get().read.format("cosmos.oltp").options(**config)
        if schema is not None:
            # noinspection PyTypeChecker
            rd = rd.schema(schema)
        else:
            rd = rd.option("spark.cosmos.read.inferSchema.enabled", "true")
        return rd.load()

    def read_table(self, table_id: str, schema: DataType = None) -> DataFrame:
        table_name = Configurator().table_name(table_id)
        return self.read_table_by_name(table_name, schema)

    def write_table_by_name(
        self, df_source: DataFrame, table_name: str, rows_per_partition: int = None
    ):
        if (
            rows_per_partition is not None
            and df_source.count() > rows_per_partition * 2
        ):
            partitions = int(1 + df_source.count() / rows_per_partition)
            df_source = df_source.repartition(partitions)
        config = self.config.copy()
        config["spark.cosmos.container"] = table_name
        (
            df_source.write.format("cosmos.oltp")
            .options(**config)
            .mode("append")  # overwrite is not supported in CosmosDB
            .save()
        )

    def write_table(
        self, df_source: DataFrame, table_id: str, rows_per_partition: int = None
    ):
        table_name = Configurator().table_name(table_id)
        self.write_table_by_name(df_source, table_name, rows_per_partition)

    def create_database(self) -> DatabaseProxy:
        """
        This method will create the database that is passed to the class init, if it
        does not exist. Also, whether it exists or not, it will create and return the
        cosmos database object that can be used to create containers under that
        database.
        """
        return self.client.create_database_if_not_exists(id=self.database)

    def create_table(
        self, table_name: str, partition_key: str, offer_throughput: int
    ) -> None:
        """
        This method will create a container(table) in the database that is passed to the
        class init. Note that, if the database does not exist, it will be created by
        this method
        """
        database = self.create_database()

        # Configure the table
        container_properties = {
            "id": table_name,
            "partition_key": PartitionKey(path=partition_key),
        }

        # Create the table
        database.create_container_if_not_exists(
            id=container_properties["id"],
            partition_key=container_properties["partition_key"],
            offer_throughput=offer_throughput,
        )

    def delete_item(
        self, table_id: str, id: Union[int, str], pk: Union[int, str] = None
    ):
        cntr = self.db_client.get_container_client(Configurator().table_name(table_id))
        cntr.delete_item(id, partition_key=pk)

    def delete_container(self, table_id: str):
        self.delete_container_by_name(Configurator().table_name(table_id))

    def delete_container_by_name(self, table_name: str):
        self.db_client.delete_container(table_name)

    def recreate_container_by_name(self, table_name: str):
        """
        Delete and recreate the container while preserving properties as
        far as possible.
        """

        for container in self.db_client.list_containers():
            if container["id"] == table_name:
                break
        else:
            raise SpetlrCosmosException(f"table not found {table_name}")

        throughput_units = (
            self.db_client.get_container_client(table_name)
            .get_throughput()
            .offer_throughput
        )

        self.db_client.delete_container(table_name)
        self.db_client.create_container(
            id=container["id"],
            partition_key=container["partitionKey"],
            offer_throughput=throughput_units,
            default_ttl=container.get("defaultTtl", None),
            indexing_policy=container.get("indexingPolicy", None),
        )

    def from_tc(self, table_id: str) -> CosmosHandle:
        tc = Configurator()
        name = tc.table_name(table_id)
        rows_per_partition = tc.get(table_id, "rows_per_partition", "")
        rows_per_partition = int(rows_per_partition) if rows_per_partition else None

        try:
            schema = SchemaManager().get_schema(table_id)
        except NoSuchSchemaException:
            schema = None

        return CosmosHandle(
            name=name,
            cosmos_db=self,
            schema=schema,
            rows_per_partition=rows_per_partition,
            partition_key=tc.get(table_id, "partition_key", None),
        )



# Defines input parameters for a cached Loader

from typing import List


class CachedLoaderParameters:
    def __init__(
        self,
        cache_table_name: str,
        key_cols: List[str],
        cache_id_cols: List[str] = None,
        *,
        do_nothing_if_more_rows_than: int = None,
        provisional_markup_step: bool = True,
        retry_cache_writes: int = 0,
        retry_cache_wait_seconds=10,
    ):
        """
        Args:
            cache_table_name: The table that holds the cache
            key_cols: the set of columns that form the primary key for a row
            cache_id_cols: These columns, added by the write operation, will be saved
                in the cache to identify e.g. the written batch.
            do_nothing_if_more_rows_than: if the input data set contains more rows than
               the specified number of rows, nothing will be written or deleted.
               Instead, the method too_many_rows() will be called.
            provisional_markup_step: if true (default) a provisional cache update
               is applied before the write or delete operation is performed, and
               rolled back afterward before applying the actual cache update.
               While this protects against data corruption, it is also slower and can
               be turned off.

        The table cache_table_name must exist and must have the following schema:
        (
            [key_cols column definitions],
            rowHash INTEGER,
            loadedTime TIMESTAMP,
            deletedTime TIMESTAMP,
            [cache_id_cols definitions, (if used)]
        )
        """

        self.cache_id_cols = cache_id_cols or []
        self.cache_table_name = cache_table_name
        self.key_cols = key_cols
        if not key_cols:
            raise ValueError("key columns must be provided")

        self.rowHash = "rowHash"
        self.loadedTime = "loadedTime"
        self.deletedTime = "deletedTime"

        self.do_nothing_if_more_rows_than = do_nothing_if_more_rows_than
        self.provisional_markup_step = provisional_markup_step
        self.retry_cache_writes = retry_cache_writes
        self.retry_cache_wait_seconds = retry_cache_wait_seconds



from unittest import TestCase

from spetlr import Configurator
from spetlr.delta import DbHandle, DeltaHandle
from spetlr.spark import Spark


class TestStopTestStreamsOnCluster(TestCase):
    """
    Verify expected setup

    """

    @classmethod
    def tearDownClass(cls) -> None:
        DbHandle.from_tc("MyDb").drop_cascade()

    def test_01_stop_test(self):
        tc = Configurator()
        tc.clear_all_configurations()
        tc.set_debug()
        tc.register("MyDb", {"name": "TestDb{ID}"})

        db = DbHandle.from_tc("MyDb")
        db.create()

        tc.register(
            "MyTbl",
            {
                "name": "TestDb{ID}.TestTbl",
            },
        )
        tbl = DeltaHandle.from_tc("MyTbl")
        tbl.create_hive_table()

        spark = Spark.get()
        catalog = spark.catalog.currentCatalog()
        print("Current catalog is", catalog)
        print(spark.sql(f"DESCRIBE DETAIL {tbl.get_tablename()}").collect()[0].asDict())



import threading
import unittest
from concurrent.futures import ThreadPoolExecutor

from spetlr import Configurator
from tests.local.configurator import tables1


class TestConfigurator(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        Configurator().clear_all_configurations()

    def test_01_import_config(self):
        tc = Configurator()
        tc.add_resource_path(tables1)

        tc.reset(debug=True)
        self.assertRegex(tc.table_name("MyFirst"), "first__.*")
        self.assertRegex(tc.table_name("MyAlias"), "first__.*")
        self.assertRegex(tc.table_name("MyForked"), "another")

        tc.reset(debug=False)
        self.assertRegex(tc.table_name("MyFirst"), "first")
        self.assertRegex(tc.table_name("MyAlias"), "first")
        self.assertRegex(tc.table_name("MyForked"), "first")

    def test_02_threading(self):
        tc = Configurator()
        tc.set_prod()

        def register_and_check(index):
            thread = threading.current_thread()
            unique = f"{index}"
            tc.register(thread.name, unique)
            assert tc.get(thread.name) == unique

        with ThreadPoolExecutor() as executor:
            executor.map(register_and_check, range(100_000))

        # initial registrations still work
        self.assertRegex(tc.table_name("MyFirst"), "first")



from typing import Dict

from spetlr.configurator.sql.types import _PeekableTokenList


def _walk_view_statement(statement: _PeekableTokenList) -> Dict:
    # for views, we only collect the name.
    name_parts = []
    while True:
        # until we managed to unpack the AS SELECT statement,
        # all the bits must be part of the name.
        # This allows unpacking names like CREATE VIEW {MYDB}.tbl{ID}
        # which will return lots of parsing errors otherwise.

        name_parts.append(next(statement).value.strip("`"))

        peeked = str(statement.peek()).strip().upper()
        if peeked.startswith(("AS", "(", "COMMENT", "TBLPROPERTIES")):
            break
        else:
            continue

    object_details = {"name": "".join(name_parts), "_is_view": True}

    return object_details



import pyspark.sql.functions as f
from pyspark.sql.types import (
    IntegerType,
    StringType,
    StructField,
    StructType,
    TimestampType,
)
from spetlrtools.testing import DataframeTestCase
from spetlrtools.time import dt_utc

from spetlr.etl.transformers.validfromto_transformer import ValidFromToTransformer
from spetlr.spark import Spark


class ValidFromToTransformerTest(DataframeTestCase):
    def test_01(self):
        max_time = dt_utc(2262, 4, 11)
        data = [
            (1, "Fender", "Telecaster", 5, dt_utc(2021, 7, 1, 10)),
            # Duplicate when comparing the primary columns (first three) and the timecol
            (1, "Fender", "Telecaster", 5, dt_utc(2021, 7, 1, 10)),
            (1, "Fender", "Telecaster", 4, dt_utc(2021, 7, 1, 11)),
            (2, "Gibson", "Les Paul", 27, dt_utc(2021, 7, 1, 11)),
            (3, "Ibanez", "RG", 22, dt_utc(2021, 8, 1, 11)),
            (3, "Ibanez", "RG", 26, dt_utc(2021, 9, 1, 11)),
            (3, "Ibanez", "RG", 18, dt_utc(2021, 10, 1, 11)),
        ]

        expected_data = [
            [
                1,
                "Fender",
                "Telecaster",
                5,
                dt_utc(2021, 7, 1, 10),
                dt_utc(2021, 7, 1, 11),
                False,
            ],
            [
                1,
                "Fender",
                "Telecaster",
                4,
                dt_utc(2021, 7, 1, 11),
                max_time,
                True,
            ],
            [
                2,
                "Gibson",
                "Les Paul",
                27,
                dt_utc(2021, 7, 1, 11),
                max_time,
                True,
            ],
            [
                3,
                "Ibanez",
                "RG",
                22,
                dt_utc(2021, 8, 1, 11),
                dt_utc(2021, 9, 1, 11),
                False,
            ],
            [
                3,
                "Ibanez",
                "RG",
                26,
                dt_utc(2021, 9, 1, 11),
                dt_utc(2021, 10, 1, 11),
                False,
            ],
            [
                3,
                "Ibanez",
                "RG",
                18,
                dt_utc(2021, 10, 1, 11),
                max_time,
                True,
            ],
        ]

        schema = StructType(
            [
                StructField("id", IntegerType(), True),
                StructField("model", StringType(), True),
                StructField("brand", StringType(), True),
                StructField("amount", IntegerType(), True),
                StructField("timecolumn", TimestampType(), True),
            ]
        )

        df = Spark.get().createDataFrame(data=data, schema=schema)

        df2 = (
            ValidFromToTransformer(
                time_col="timecolumn", wnd_cols=["id", "model", "brand"]
            )
            .process(df)
            .drop("timecolumn")
            .orderBy(f.col("ValidFrom").asc(), f.col("ValidTo").asc())
        )

        self.assertDataframeMatches(
            df2,
            None,
            expected_data=expected_data,
        )



"""
A python SPark ETL libRary (SPETLR) for Databricks.
See https://github.com/spetlr-org/spetlr for details
"""

from spetlr import etl, functions, spark, sql  # noqa: F401
from spetlr.configurator.configurator import Configurator  # noqa: F401

from .version import __version__  # noqa: F401

DEBUG = False


def dbg(*args, **kwargs):
    if DEBUG:
        print(*args, **kwargs)



import unittest
from datetime import datetime as dt
from datetime import timezone

from spetlr.eh.EventHubCaptureExtractor import EventHubCaptureExtractor

utc = timezone.utc


class EventHubCaptureExtractorTests(unittest.TestCase):
    def test_breakup(self):
        eh = EventHubCaptureExtractor(
            path="/does/not/matter/since/unused", partitioning="ymdh"
        )
        parts = eh._break_into_partitioning_parts(
            dt(2020, 11, 30, 22, tzinfo=utc), dt(2022, 5, 3, 3, tzinfo=utc)
        )
        self.assertEqual(
            parts,
            [
                "y=2020/m=11/d=30/h=22/",
                "y=2020/m=11/d=30/h=23/",
                "y=2020/m=12/",
                "y=2021/",
                "y=2022/m=01/",
                "y=2022/m=02/",
                "y=2022/m=03/",
                "y=2022/m=04/",
                "y=2022/m=05/d=01/",
                "y=2022/m=05/d=02/",
                "y=2022/m=05/d=03/h=00/",
                "y=2022/m=05/d=03/h=01/",
                "y=2022/m=05/d=03/h=02/",
            ],
        )

    def test_open_interval(self):
        eh = EventHubCaptureExtractor(
            path="/does/not/matter/since/unused", partitioning="ymdh"
        )
        eh._now_utc = lambda: dt(2022, 5, 3, 3, tzinfo=utc)
        fp, tp = eh._validated_slice_arguments(dt(2020, 11, 30, 22, tzinfo=utc), None)

        self.assertEqual(tp, dt(2023, 1, 1, tzinfo=utc))

        parts = eh._break_into_partitioning_parts(fp, tp)
        self.assertEqual(
            parts,
            [
                "y=2020/m=11/d=30/h=22/",
                "y=2020/m=11/d=30/h=23/",
                "y=2020/m=12/",
                "y=2021/",
                "y=2022/",
            ],
        )



import unittest

from spetlr.db_auto import getDbApi


class ApiTests(unittest.TestCase):
    def test_01_configureApi(self):
        getDbApi()

    def test_02_make_a_call(self):
        db = getDbApi()

        self.assertIsNotNone(db.workspace.get_status("/Shared"))



import pyspark.sql.types as T
from spetlrtools.testing import DataframeTestCase

from spetlr.etl.transformers import DropColumnsTransformer
from spetlr.spark import Spark


class TestDropColumnsTransformer(DataframeTestCase):
    def test_drop_columns_transformer(self):
        inputSchema = T.StructType(
            [
                T.StructField("id", T.LongType(), True),
                T.StructField("text1", T.StringType(), True),
                T.StructField("text2", T.StringType(), True),
            ]
        )

        inputData = [
            (
                1,
                "text1",
                "text2",
            ),
        ]

        df_input = Spark.get().createDataFrame(data=inputData, schema=inputSchema)

        expectedData = [
            (1,),
        ]

        df_transformed = DropColumnsTransformer(columnList=["text1", "text2"]).process(
            df_input
        )
        self.assertDataframeMatches(df_transformed, None, expectedData)



from collections import OrderedDict
from typing import Any, List, Tuple, Union

from pyspark.sql import DataFrame
from pyspark.sql.types import ArrayType, Row, StructType

from spetlr.spark import Spark

_column_selection = Union[str, Tuple[str, List["_column_selection"]]]


class DataframeCreator:
    @classmethod
    def make_partial(
        cls, schema: StructType, columns: List[_column_selection], data: List[Tuple]
    ) -> DataFrame:
        rows = []
        if data:
            for data_row in data:
                rows.append(cls._make_row(schema, data_row, columns))
        return Spark.get().createDataFrame(rows, schema=schema)

    @classmethod
    def make(cls, schema: StructType, data: List[Tuple]) -> DataFrame:
        return Spark.get().createDataFrame(data, schema=schema)

    @classmethod
    def _make_row_array(
        cls, schema: ArrayType, data_row: List[Any], columns: List[_column_selection]
    ) -> List[Any]:
        element = schema.elementType
        if isinstance(element, ArrayType):
            return [cls._make_row_array(element, item, columns) for item in data_row]
        elif isinstance(element, StructType):
            return [cls._make_row(element, item, columns) for item in data_row]
        else:
            return data_row

    @classmethod
    def _make_row(
        cls, schema: StructType, data_row: Tuple, columns: List[_column_selection]
    ) -> Row:
        schema_dict = {f.name: f.dataType for f in schema.fields}
        in_data = dict()
        for column, item in zip(columns, data_row):
            if isinstance(column, str):
                # we are dealing with a simple column
                assert column in schema.names, f"column '{column}' is missing!"
                in_data[column] = item

            else:
                # we are dealing with a composite column
                # here we need to recurse so that we can set the data to a Row object
                column_name = column[0]
                column_schema_selection = column[1]
                column_schema = schema_dict[column_name]
                if isinstance(column_schema, StructType):
                    in_data[column_name] = cls._make_row(
                        column_schema, item, column_schema_selection
                    )
                else:
                    assert isinstance(
                        column_schema, ArrayType
                    ), f"column type is {type(column_schema)}!"
                    in_data[column_name] = cls._make_row_array(
                        column_schema, item, column_schema_selection
                    )

                    # fix the order of the items in the row item
                    # while also given the row elements keys
        kwargs = OrderedDict()
        for column in schema.names:
            if column in in_data:
                kwargs[column] = in_data[column]
            else:
                kwargs[column] = None

        return Row(**kwargs)



from typing import List

from sqlparse import engine, keywords, tokens
from sqlparse.engine import grouping
from sqlparse.lexer import Lexer
from sqlparse.sql import Statement

lex = Lexer.get_default_instance()
my_regex = [
    (r"IF\s+NOT\s+EXISTS\b", tokens.Keyword),
    (r"ZORDER\s+BY\b", tokens.Keyword),
    (r"PARTITIONED\s+BY\b", tokens.Keyword),
    (r"SORTED\s+BY\b", tokens.Keyword),
    (r"CLUSTER\s+BY\b", tokens.Keyword),
    (r"CLUSTERED\s+BY\b", tokens.Keyword),
    (r"WITH\s+DBPROPERTIES\b", tokens.Keyword),
    (r"BLOOMFILTER\s+INDEX\b", tokens.Keyword),
    (r"(DEEP|SHALLOW)\s+CLONE\b", tokens.Keyword),
    (r"(MSCK|FSCK)\s+REPAIR\b", tokens.Keyword),
    (r"[<>=~!]", tokens.Operator.Comparison),  # avoid >> being parsed as one token
]

KEYWORDS_DBX = {
    "BLOOMFILTER": tokens.Keyword,
    "BUCKETS": tokens.Keyword,
    "DBPROPERTIES": tokens.Keyword,
    "DETAIL": tokens.Keyword,
    "HISTORY": tokens.Keyword,
    "METADATA": tokens.Keyword,
    "MSCK": tokens.Keyword,
    "OPTIMIZE": tokens.Keyword,
    "PARTITIONS": tokens.Keyword,
    "REFRESH": tokens.Keyword,
    "REPAIR": tokens.Keyword,
    "SYNC": tokens.Keyword,
    "VACUUM": tokens.Keyword,
    "ZORDER": tokens.Keyword,
}


lex.clear()

lex.set_SQL_REGEX(keywords.SQL_REGEX[:38] + my_regex + keywords.SQL_REGEX[38:])
lex.add_keywords(keywords.KEYWORDS_COMMON)
lex.add_keywords(keywords.KEYWORDS_ORACLE)
lex.add_keywords(keywords.KEYWORDS_PLPGSQL)
lex.add_keywords(keywords.KEYWORDS_HQL)
lex.add_keywords(keywords.KEYWORDS_MSACCESS)
lex.add_keywords(keywords.KEYWORDS)

lex.add_keywords(KEYWORDS_DBX)


def mygrouping(stmt):
    for func in [
        # group_comments,
        # _group_matching
        grouping.group_brackets,
        grouping.group_parenthesis,
        grouping.group_case,
        grouping.group_if,
        grouping.group_for,
        grouping.group_begin,
        grouping.group_functions,
        grouping.group_where,
        grouping.group_period,
        grouping.group_arrays,
        grouping.group_identifier,
        grouping.group_order,
        grouping.group_typecasts,
        grouping.group_tzcasts,
        grouping.group_typed_literal,
        grouping.group_operator,
        grouping.group_comparison,
        grouping.group_as,
        grouping.group_aliased,
        grouping.group_assignment,
        grouping.align_comments,
        grouping.group_identifier_list,
        grouping.group_values,
    ]:
        func(stmt)
    return stmt


def parsestream(stream, encoding=None):
    stack = engine.FilterStack()
    # stack.enable_grouping()
    for stmt in stack.run(stream, encoding):
        mygrouping(stmt)
        yield stmt


def parse(sql: str, encoding=None) -> List[Statement]:
    return list(parsestream(sql, encoding))



from unittest.mock import Mock, patch

from pyspark.sql import DataFrame
from pyspark.sql import functions as f
from spetlrtools.testing import DataframeTestCase

from spetlr.etl import Transformer
from spetlr.etl.loaders import SimpleLoader
from spetlr.orchestrators import EhToDeltaBronzeOrchestrator
from spetlr.orchestrators.eh2bronze.EhToDeltaBronzeTransformer import (
    EhToDeltaBronzeTransformer,
)
from spetlr.orchestrators.ehjson2delta.EhJsonToDeltaExtractor import (
    EhJsonToDeltaExtractor,
)
from spetlr.spark import Spark


class EhToDeltaBronzeOrchestratorTests(DataframeTestCase):
    def test_01_default(self):
        empty_df = Spark.get().createDataFrame([], schema="Id INTEGER")
        eh_mock = Mock()
        dh_mock = Mock()

        eh_mock.read = Mock(return_value=empty_df)

        with patch.object(EhJsonToDeltaExtractor, "read", return_value=empty_df) as p1:
            with patch.object(SimpleLoader, "save", return_value=empty_df) as p2:
                with patch.object(
                    EhToDeltaBronzeTransformer, "process", return_value=empty_df
                ) as p3:
                    orchestrator = EhToDeltaBronzeOrchestrator(eh=eh_mock, dh=dh_mock)
                    orchestrator.execute()

                    # EhJsonToDeltaExtractor should be called once
                    p1.assert_called_once()
                    # SimpleLoader should be called once
                    p2.assert_called_once()
                    # The transformer should be called once
                    p3.assert_called_once()

    def test_02_w_transformer(self):
        empty_df = Spark.get().createDataFrame([], schema="Id INTEGER")
        eh_mock = Mock()
        dh_mock = Mock()

        class TestFilter(Transformer):
            def process(self, df: DataFrame) -> DataFrame:
                return df.withColumn("Test", f.lit("Test"))

        class TestOrchestrator(EhToDeltaBronzeOrchestrator):
            def __init__(self):
                super().__init__(
                    eh=eh_mock,
                    dh=dh_mock,
                )
                self.filter_with(TestFilter())

        eh_mock.read = Mock(return_value=empty_df)

        with patch.object(EhJsonToDeltaExtractor, "read", return_value=empty_df) as p1:
            with patch.object(SimpleLoader, "save", return_value=empty_df) as p2:
                with patch.object(
                    EhToDeltaBronzeTransformer, "process", return_value=empty_df
                ) as p3:
                    with patch.object(
                        TestFilter, "process", return_value=empty_df
                    ) as p4:
                        orchestrator = TestOrchestrator()
                        orchestrator.execute()

                        # EhJsonToDeltaExtractor should be called once
                        p1.assert_called_once()
                        # SimpleLoader should be called once
                        p2.assert_called_once()
                        # The transformer should be called once
                        p3.assert_called_once()
                        # The test transformer should be called once
                        p4.assert_called_once()



from . import SpetlrException


class SpetlrConfiguratorException(SpetlrException):
    pass


class SpetlrConfiguratorInvalidSqlException(SpetlrConfiguratorException):
    pass


class SpetlrConfiguratorInvalidSqlCommentsException(SpetlrConfiguratorException):
    pass


class DeprecationException(SpetlrException):
    pass



from datetime import datetime
from typing import Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from dateutil.parser import parse
from pandas.core.generic import NDFrameT
from pyspark.sql import DataFrame
from pytz import timezone, utc

from spetlr.power_bi.PowerBiException import PowerBiException
from spetlr.spark import Spark


class SparkPandasDataFrame:
    # converts data types from Spark to Pandas and vice versa
    SPARK_TO_PANDAS_TYPES = {
        "long": "Int64",
        "int": "Int64",
        "timestamp": "datetime64[ns]",
        "boolean": "bool",
        "string": "object",
        "default": "object",
    }

    def __init__(
        self,
        json: Union[Dict, List],
        schema: List[Tuple[Union[str, Callable[[pd.DataFrame], NDFrameT]], str, str]],
        *,
        indexing_columns: Optional[Union[int, List[int], str, List[str]]] = None,
        sorting_columns: Optional[Union[int, List[int], str, List[str]]] = None,
        ascending: bool = True,
        local_timezone_name: str = None,
    ):
        """
        Provides a class for loading a JSON into Pandas, and for converting
        between Pandas and Spark.

        :param dict json: Data source, which is a dictionary representing a JSON.
        :param dict[tuple[str/callable, str, str]: schema definition where the first
            value in the tuple is the original column name in the JSON, the second
            is the desired column name in the output, and the third is the Spark type.
            The first value can also be a lambda expression taking a Pandas
            data frame as input and calculating the column value from
            other available columns, e.g. lambda df: df["endTime"] - df["startTime"]
        :param int/str indexing_columns: indexes (within the schema) or names
            of the indexing columns.
        :param int/str sorting_columns: indexes (within the schema) or names
            of the sorting columns.
        :param str local_timezone_name: The timezone to use when parsing
            timestamp columns. The default timezone is UTC.
            If the timezone is UTC, all timestamp columns will have a suffix "Utc".
            Otherwise they will have a suffix "Local".
        """

        self.local_timezone_name = local_timezone_name or "UTC"
        self.is_utc = self.local_timezone_name.upper() == "UTC"
        self.time_column_suffix = "Utc" if self.is_utc else "Local"
        self.schema = schema

        for i, column in enumerate(schema):
            if column[2] == "timestamp":
                schema[i] = (column[0], column[1] + self.time_column_suffix, column[2])
        df = pd.DataFrame(
            json,
            columns=[column[0] for column in schema if isinstance(column[0], str)],
        )
        if df.empty:
            self.df = pd.DataFrame()
        else:
            df = df.replace(np.nan, None)
            if indexing_columns is not None:
                df.set_index(self._get_columns(indexing_columns, False))
            for column in schema:
                if isinstance(column[0], str):
                    if column[2] == "timestamp":
                        df[column[0]] = df[column[0]].apply(
                            lambda text: self._parse_time(text)
                        )
                    else:
                        df[column[0]] = df[column[0]].astype(
                            self.SPARK_TO_PANDAS_TYPES.get(
                                column[2], self.SPARK_TO_PANDAS_TYPES["default"]
                            )
                        )
            for i, column in enumerate(schema):
                if callable(column[0]):
                    df.insert(
                        i,
                        column[1],
                        column[0](df).astype(
                            self.SPARK_TO_PANDAS_TYPES.get(
                                column[2], self.SPARK_TO_PANDAS_TYPES["default"]
                            )
                        ),
                    )
            for column in schema:
                if column[2] == "timestamp" and isinstance(column[0], str):
                    df[column[0]] = df[column[0]].apply(
                        lambda text: self._localize_time(text, self.local_timezone_name)
                    )
            df = df.rename(
                columns=dict(
                    (column[0], column[1])
                    for column in schema
                    if isinstance(column[0], str)
                ),
            )
            if sorting_columns is not None:
                df = df.sort_values(
                    self._get_columns(sorting_columns, True), ascending=ascending
                )
            self.df = df

    def _get_columns(
        self, identifiers: Union[int, List[int], str, List[str]], target: bool
    ) -> Union[str, List[str]]:
        """
        Returns the name of the specified column in the schema.

        :param int/str identifiers: the column(s) to look for, which is either
            the column index(es), or column source name(s), or column target name(s).
        :param bool target: return the target column name if target==True,
            or source column name otherwise.
        :return: the requested column
        :rtype: str
        :raises PowerBiException: if the column cannot be found
        """

        result = []
        if isinstance(identifiers, int) or isinstance(identifiers, str):
            identifiers = [identifiers]
        if isinstance(identifiers, list):
            for identifier in identifiers:
                if isinstance(identifier, int) and identifier < len(self.schema):
                    result.append(self.schema[identifier][int(target)])
                elif isinstance(identifier, str):
                    result.append(
                        next(
                            (
                                column[int(target)]
                                for column in self.schema
                                if identifier == column[0] or identifier == column[1]
                            ),
                            None,
                        )
                    )
        if not result or (
            isinstance(identifiers, list) and len(result) != len(identifiers)
        ):
            raise PowerBiException(f"{identifiers} column(s) not found in the schema")
        return result if len(result) > 1 else result[0]

    @staticmethod
    def _parse_time(text: str) -> Optional[datetime]:
        """
        Parses an ISO 8601 time text and converts it to datetime.
        The result is a timezone-unaware UTC datetime regardless of input.

        :param str text: an ISO 8601 time text or None
        :return: timezone-unaware UTC datetime or None
        :rtype: datetime
        """

        if text is None or text != text:
            return None
        time = parse(text)
        if time.tzinfo is None or time.tzinfo.utcoffset(time) is None:
            time = utc.localize(time)
        else:
            time = time.astimezone(utc)
        return time.replace(microsecond=0, tzinfo=None)

    @staticmethod
    def _localize_time(
        time: Optional[datetime], local_timezone_name: str
    ) -> Optional[datetime]:
        """
        Converts a timezone-unaware UTC datetime to a timezone-unaware local datetime.

        :param datetime time: timezone-unaware UTC datetime or None
        :return: timezone-unaware local datetime or None
        :rtype: datetime
        """

        if time is None or time != time:
            return None
        if time.tzinfo is None or time.tzinfo.utcoffset(time) is None:
            time = utc.localize(time)
        time = time.astimezone(timezone(local_timezone_name))
        return time.replace(tzinfo=None)

    def get_pandas_df(self) -> Optional[pd.DataFrame]:
        """
        Returns the data frame as a Pandas data frame.

        :return: Pandas data frame
        :rtype: Pandas data frame
        """

        return self.df

    def get_spark_df(self) -> Optional[DataFrame]:
        """
        Returns the data frame as a Spark data frame.

        :return: Spark data frame
        :rtype: Spark data frame
        """

        df = self.df
        if df is None:
            return None
        spark_schema = ", ".join(f"{column[1]} {column[2]}" for column in self.schema)
        if df.empty:
            return Spark.get().createDataFrame(df, spark_schema)
        else:
            return Spark.get().createDataFrame(df)

    def show(
        self,
        message: str,
        when_empty: str,
        *,
        filter_columns: Optional[List[Tuple[str, str]]] = None,
    ) -> None:
        """
        Displays the Pandas data frame.

        :param str message: The message to show when the data frame is not empty.
        :param str when_empty: The message to show when the data frame is empty.
           Note: Pandas fails when showing empty data frames.
        :param list[tuple[str, str]] filter_columns: a subset of columns to show
            and their new names
        :rtype: None
        """

        if self.df is None or self.df.empty:
            print(when_empty)
        else:
            print(message)
            df = self.df
            if filter_columns:
                df = df.rename(columns=dict(filter_columns))
                df = df[[column[1] for column in filter_columns]]
            df.display()

    def get_local_time_str(self, time: datetime) -> Optional[str]:
        """
        Returns the specified time as a local timestamp converted to text.

        :param datetime time: the time to convert from
        :return: text of a timezone-aware local datetime or None
        :rtype: str
        """

        time_description_suffix = " (UTC)" if self.is_utc else " (local time)"
        if time is not None:
            return time.strftime("%Y-%m-%d %H:%M") + time_description_suffix
        return None

    def get_utc_time(self, time: datetime) -> Optional[datetime]:
        """
        Returns the specified time as an UTC timestamp.

        :param datetime time: the time to convert from
        :return: a timezone-aware UTC datetime or None
        :rtype: datetime
        """

        zone = timezone(self.local_timezone_name)
        if time is not None:
            if self.is_utc:
                time = utc.localize(time)
            else:
                time = zone.localize(time).astimezone(utc)
        return time

    def append(
        self,
        source,
        prefix_columns: List[Tuple[str, str, str]],
        suffix_columns: List[Tuple[str, str, str]],
    ):
        """
        Appends another instance of this class to this data frame, which will result
        in a union of two Pandas data frames. Additionally, new columns should
        be added to this instance, which are already present in the other instance.
        This will allow to make a distinction between both data frames.

        :param SparkPandasDataFrame source: another object of this class
            that needs to be appended to this data frame
        :param list[tuple[str, str]] prefix_columns: new columns with values
            to add at the beginning
        :param list[tuple[str, str]] suffix_columns: new columns with values
            to add at the end
        :rtype: SparkPandasDataFrame
        """
        for column in reversed(prefix_columns):
            self.schema.insert(0, (column[0], column[0], column[2]))
            if not self.df.empty:
                self.df.insert(0, column[0], column[1])
        for column in suffix_columns:
            self.schema.append((column[0], column[0], column[2]))
            if not self.df.empty:
                self.df[column[0]] = column[1]
        if source is not None and not source.get_pandas_df().empty:
            self.df = pd.concat([source.get_pandas_df(), self.df])
        if not self.df.empty:
            self.df = self.df.reset_index(drop=True)
        return self



# These tests will be activated when the corresponding cluster tests get fixed

# from unittest import TestCase

# from spetlr.deltaspec.exceptions import InvalidTableName
# from spetlr.deltaspec.helpers import TableName


# class TestTableName(TestCase):
#     def test_01_levels(self):
#         n = TableName.from_str(None)
#         self.assertEqual(n.level(), 0)
#         self.assertEqual(str(n), "")
#         self.assertEqual(TableName.from_str(None), TableName())
#         self.assertEqual(TableName.from_str(""), TableName())

#         t = TableName.from_str("tbl")
#         self.assertEqual(t.level(), 1)
#         self.assertEqual(str(t), "tbl")

#         d = TableName.from_str("db.tbl")
#         self.assertEqual(d.level(), 2)
#         self.assertEqual(str(d), "db.tbl")

#         c = TableName.from_str("ctlg.db.tbl")
#         self.assertEqual(c.level(), 3)
#         self.assertEqual(str(c), "ctlg.db.tbl")

#         with self.assertRaises(InvalidTableName):
#             TableName(schema="db")

#         with self.assertRaises(InvalidTableName):
#             TableName(catalog="c", schema="db")

#         self.assertEqual(str(c.to_level(2)), "db.tbl")
#         self.assertEqual(str(c.to_level(1)), "tbl")
#         self.assertEqual(str(c.to_level(0)), "")



import warnings
from typing import List

from .types import EtlBase, dataset_group


class Orchestrator(EtlBase):
    """
    It is up to the user of this library that extractors,
    transformers and loaders live up to their names and are not
    used in a wrong order.
    """

    def __init__(self, suppress_composition_warning=False):
        super().__init__()
        self.steps: List[EtlBase] = []
        self.suppress_composition_warning = suppress_composition_warning

    def step(self, etl: EtlBase) -> "Orchestrator":
        self.steps.append(etl)
        return self

    # these are just synonyms for readability
    extract_from = step
    transform_with = step
    load_into = step

    def etl(self, inputs: dataset_group = None) -> dataset_group:
        inputs = inputs or {}

        # make a shallow copy of the inputs for waring after the first step
        datasets = inputs.copy()
        if not self.steps:
            raise NotImplementedError("The orchestrator has no steps.")

        # treat the fist step differently to warn in case the input was not handled
        datasets = self.steps[0].etl(datasets)
        if len(inputs) and (len(inputs) + 1 == len(datasets)):
            # There were inputs to the orchestrator,
            # and the first step did not clean them up.
            # expect problems.

            # let's double check:
            if (
                # step 1 has retained all input keys
                set(inputs.keys()).issubset(set(datasets.keys()))
                and
                # and has not changed their values
                all(id(inputs[k]) == id(datasets[k]) for k in inputs.keys())
                and not self.suppress_composition_warning
            ):
                warnings.warn(
                    "You used inputs to the orchestrator, "
                    "and the first step did not make use of them. "
                    "Expect problems in your etl pipeline. To avoid this, "
                    "write extractors that clean up in self.previous_extractions"
                )

        for step in self.steps[1:]:
            datasets = step.etl(datasets)
        return datasets

    execute = etl



import unittest
from datetime import datetime

from spetlr.extractors import (
    EventhubStreamExtractor,
    InvalidEventhubStreamExtractorParameters,
)


class EventhubStreamExtractorTest(unittest.TestCase):
    def test_create_with_connectionString(self):
        eventhubStreamExtractor = EventhubStreamExtractor(
            connectionString="testConnectionString",
            consumerGroup="testConsumerGroup",
            maxEventsPerTrigger=100000,
        )

        self.assertEqual(
            eventhubStreamExtractor.connectionString, "testConnectionString"
        )

    def test_create_without_connectionString(self):
        eventhubStreamExtractor = EventhubStreamExtractor(
            namespace="testNamespace",
            eventhub="testEventhub",
            accessKeyName="testAccessKeyName",
            accessKey="testAccessKey",
            consumerGroup="testConsumerGroup",
            maxEventsPerTrigger=100000,
        )

        self.assertEqual(
            eventhubStreamExtractor.connectionString,
            "Endpoint=sb://testNamespace.servicebus.windows.net/testEventhub;"
            "EntityPath=testEventhub;SharedAccessKeyName=testAccessKeyName;"
            "SharedAccessKey=testAccessKey",
        )

    def test_raise_create_exeption(self):
        with self.assertRaises(InvalidEventhubStreamExtractorParameters):
            EventhubStreamExtractor(
                connectionString=None,
                namespace=None,
                eventhub="testEventhub",
                accessKeyName="testAccessKeyName",
                accessKey="testAccessKey",
                consumerGroup="testConsumerGroup",
                maxEventsPerTrigger=100000,
            )

        with self.assertRaises(InvalidEventhubStreamExtractorParameters):
            EventhubStreamExtractor(
                connectionString=None,
                namespace="testNamespace",
                eventhub=None,
                accessKeyName="testAccessKeyName",
                accessKey="testAccessKey",
                consumerGroup="testConsumerGroup",
                maxEventsPerTrigger=100000,
            )

        with self.assertRaises(InvalidEventhubStreamExtractorParameters):
            EventhubStreamExtractor(
                connectionString=None,
                namespace="testNamespace",
                eventhub="testEventhub",
                accessKeyName=None,
                accessKey="testAccessKey",
                consumerGroup="testConsumerGroup",
                maxEventsPerTrigger=100000,
            )

        with self.assertRaises(InvalidEventhubStreamExtractorParameters):
            EventhubStreamExtractor(
                connectionString=None,
                namespace="testNamespace",
                eventhub="testEventhub",
                accessKeyName="testAccessKeyName",
                accessKey=None,
                consumerGroup="testConsumerGroup",
                maxEventsPerTrigger=100000,
            )

    def test_create_start_from_beginning_of_stream(self):
        eventhubStreamExtractor = EventhubStreamExtractor(
            connectionString="testConnectionString",
            consumerGroup="testConsumerGroup",
            maxEventsPerTrigger=100000,
        )

        expectedStartingEventPosition = {
            "offset": "-1",  # Start stream from beginning
            "seqNo": -1,  # not in use
            "enqueuedTime": None,  # not in use
            "isInclusive": True,
        }

        self.assertEqual(
            eventhubStreamExtractor.startingEventPosition, expectedStartingEventPosition
        )

    def test_create_start_from_timestamp(self):
        timestampToUse = datetime.utcnow()

        eventhubStreamExtractor = EventhubStreamExtractor(
            connectionString="testConnectionString",
            consumerGroup="testConsumerGroup",
            maxEventsPerTrigger=100000,
            startEnqueuedTime=timestampToUse,
        )

        expectedStartingEventPosition = {
            "offset": None,  # not in use
            "seqNo": -1,  # not in use
            "enqueuedTime": timestampToUse.strftime(
                "%Y-%m-%dT%H:%M:%S.%fZ"
            ),  # Start from timestamp
            "isInclusive": True,
        }

        self.assertEqual(
            eventhubStreamExtractor.startingEventPosition, expectedStartingEventPosition
        )

    def test_read_eventhub_stream(self):
        # Streaming is not testable at the moment
        pass


if __name__ == "__main__":
    unittest.main()



# These tests will be activated when the corresponding cluster tests get fixed

# import dataclasses
# import unittest
# from textwrap import dedent

# from pyspark.sql import types as t

# from spetlr import Configurator
# from spetlr.deltaspec import DeltaTableSpecDifference, TableSpectNotEnforcable
# from spetlr.deltaspec.DeltaTableSpec import DeltaTableSpec
# from tests.cluster.delta.deltaspec import tables
# from tests.local.delta import sql


# class TestDeltaTableSpec(unittest.TestCase):
#     @classmethod
#     def setUpClass(cls) -> None:
#         c = Configurator()
#         c.clear_all_configurations()
#         c.set_prod()

#         cls.base = tables.base
#         cls.target = tables.target

#     def test_01_create_statements(self):
#         self.assertEqual(
#             self.base.get_sql_create(),
#             dedent(
#                 """\
#                 CREATE TABLE myDeltaTableSpecTestDb{ID}.tbl
#                 (
#                   c double,
#                   d string COMMENT "Whatsupp",
#                   onlyb int,
#                   a int,
#                   b int
#                 )
#                 USING DELTA
#                 LOCATION "dbfs:/tmp/somewhere{ID}/over/the/rainbow"
#                 TBLPROPERTIES (
#                   "delta.columnMapping.mode" = "name",
#                   "delta.minReaderVersion" = "2",
#                   "delta.minWriterVersion" = "5"
#                 )
#                 """
#             ),
#         )

#     def test_02_alter_statements_raise(self):
#         Configurator().set_prod()
#         forward_diff = self.target.compare_to(self.base)
#         with self.assertRaises(TableSpectNotEnforcable):
#             forward_diff.alter_statements(
#                 allow_columns_add=False,
#                 allow_columns_type_change=True,
#                 allow_columns_drop=True,
#                 allow_columns_reorder=True,
#             )

#         with self.assertRaises(TableSpectNotEnforcable):
#             forward_diff.alter_statements(
#                 allow_columns_add=True,
#                 allow_columns_type_change=False,
#                 allow_columns_drop=True,
#                 allow_columns_reorder=True,
#             )

#         with self.assertRaises(TableSpectNotEnforcable):
#             forward_diff.alter_statements(
#                 allow_columns_add=True,
#                 allow_columns_type_change=True,
#                 allow_columns_drop=True,
#                 allow_columns_reorder=False,
#             )
#         print("No raise when they are all warnings")
#         forward_diff.alter_statements(errors_as_warnings=True)

#     def test_03_diff_alter_statements(self):
#         Configurator().set_prod()
#         forward_diff = self.target.compare_to(self.base)

#         self.assertEqual(
#             forward_diff.alter_statements(
#                 allow_columns_add=True,
#                 allow_columns_type_change=True,
#                 allow_columns_drop=True,
#                 allow_columns_reorder=True,
#             ),
#             [
#                 "ALTER TABLE mydeltatablespectestdb.tbl "
#                 'SET TBLPROPERTIES ("my.cool.peoperty" = "bacon")',
#                 "ALTER TABLE mydeltatablespectestdb.tbl DROP COLUMNS (b, onlyb)",
#                 dedent(
#                     """\
#                 ALTER TABLE mydeltatablespectestdb.tbl ADD COLUMNS (
#                   b string,
#                   onlyt string COMMENT "Only in target"
#                 )"""
#                 ),
#                 # "ALTER TABLE mydeltatablespectestdb.tbl "
#                 #  "ALTER COLUMN a DROP NOT NULL",
#                 # "ALTER TABLE mydeltatablespectestdb.tbl "
#                 #  "ALTER COLUMN d SET NOT NULL",
#                 "ALTER TABLE mydeltatablespectestdb.tbl ALTER COLUMN a COMMENT"
#                 ' "gains not null"',
#                 'ALTER TABLE mydeltatablespectestdb.tbl ALTER COLUMN d COMMENT ""',
#                 "ALTER TABLE mydeltatablespectestdb.tbl ALTER COLUMN a FIRST",
#                 "ALTER TABLE mydeltatablespectestdb.tbl ALTER COLUMN b AFTER a",
#                 "ALTER TABLE mydeltatablespectestdb.tbl ALTER COLUMN onlyt AFTER d",
#                 'COMMENT ON TABLE mydeltatablespectestdb.tbl '
#                 'is "Contains useful data"',
#             ],
#         )

#         reverse_diff = self.base.compare_to(self.target)
#         self.assertEqual(
#             reverse_diff.alter_statements(
#                 allow_columns_add=True,
#                 allow_columns_type_change=True,
#                 allow_columns_drop=True,
#                 allow_columns_reorder=True,
#                 allow_name_change=True,
#                 allow_location_change=True,
#             ),
#             [
#                 "ALTER TABLE mydeltatablespectestdb.tbl "
#                 'UNSET TBLPROPERTIES ("my.cool.peoperty")',
#                 "ALTER TABLE mydeltatablespectestdb.tbl DROP COLUMNS (b, onlyt)",
#                 dedent(
#                     """\
#                 ALTER TABLE mydeltatablespectestdb.tbl ADD COLUMNS (
#                   onlyb int,
#                   b int
#                 )"""
#                 ),
#                 # "ALTER TABLE mydeltatablespectestdb.tbl "
#                 # "ALTER COLUMN d DROP NOT NULL",
#                 # "ALTER TABLE mydeltatablespectestdb.tbl "
#                 #  "ALTER COLUMN a SET NOT NULL",
#                 "ALTER TABLE mydeltatablespectestdb.tbl ALTER COLUMN d COMMENT "
#                 '"Whatsupp"',
#                 'ALTER TABLE mydeltatablespectestdb.tbl ALTER COLUMN a COMMENT ""',
#                 "ALTER TABLE mydeltatablespectestdb.tbl ALTER COLUMN c FIRST",
#                 "ALTER TABLE mydeltatablespectestdb.tbl ALTER COLUMN d AFTER c",
#                 "ALTER TABLE mydeltatablespectestdb.tbl ALTER COLUMN onlyb AFTER d",
#                 "COMMENT ON TABLE mydeltatablespectestdb.tbl is null",
#             ],
#         )

#     def test_04_check_init_protections(self):
#         tbl1 = DeltaTableSpec(
#             name="myDeltaTableSpecTestDb{ID}.tbl",
#             schema=t.StructType(
#                 fields=[
#                     t.StructField(name="c", dataType=t.DoubleType()),
#                     t.StructField(
#                         name="d",
#                         dataType=t.StringType(),
#                         nullable=False,
#                         metadata={"comment": "Whatsupp"},
#                     ),
#                     t.StructField(name="onlyb", dataType=t.IntegerType()),
#                     t.StructField(name="a", dataType=t.IntegerType()),
#                     t.StructField(name="b", dataType=t.StringType()),
#                 ]
#             ),
#             location="/somewhere/over/the{ID}/rainbow",
#         )
#         tbl2 = DeltaTableSpec(
#             name="spark_catalog.mydeltatablespectestdb.tbl",
#             schema=t.StructType(
#                 fields=[
#                     t.StructField(name="c", dataType=t.DoubleType()),
#                     t.StructField(
#                         name="d",
#                         dataType=t.StringType(),
#                         nullable=False,
#                         metadata={"comment": "Whatsupp"},
#                     ),
#                     t.StructField(name="onlyb", dataType=t.IntegerType()),
#                     t.StructField(name="a", dataType=t.IntegerType()),
#                     t.StructField(name="b", dataType=t.StringType()),
#                 ]
#             ),
#             location="dbfs:/somewhere/over/the/rainbow",
#         )
#         Configurator().set_prod()
#         d = tbl2.compare_to(tbl1.fully_substituted())
#         self.assertFalse(d.is_different(), d)

#     def test_05_namechagne(self):
#         statements = tables.newname.compare_to(tables.oldname).alter_statements(
#             allow_columns_add=True,
#             allow_columns_type_change=True,
#             allow_columns_drop=True,
#             allow_columns_reorder=True,
#             allow_name_change=True,
#             allow_location_change=True,
#         )
#         self.assertEqual(
#             statements,
#             [
#                 "ALTER TABLE mydeltatablespectestdb.namechange_old RENAME TO "
#                 "mydeltatablespectestdb.namechange_new"
#             ],
#         )

#     def test_06_location_change(self):
#         statements = tables.newlocation.compare_to(
#                   tables.oldlocation).alter_statements(
#             allow_columns_add=True,
#             allow_columns_type_change=True,
#             allow_columns_drop=True,
#             allow_columns_reorder=True,
#             allow_name_change=True,
#             allow_location_change=True,
#         )
#         self.assertEqual(
#             statements,
#             [
#                 "CREATE TABLE delta.`dbfs:/tmp/somewhere/locchange/new`\n"
#                 "(\n"
#                 "  b string,\n"
#                 "  c double,\n"
#                 "  d string\n"
#                 ")\n"
#                 "USING DELTA\n"
#                 'LOCATION "dbfs:/tmp/somewhere/locchange/new"\n'
#                 'COMMENT "Contains useful data"\n'
#                 "TBLPROPERTIES (\n"
#                 '  "delta.columnMapping.mode" = "name",\n'
#                 '  "delta.minReaderVersion" = "2",\n'
#                 '  "delta.minWriterVersion" = "5"\n'
#                 ")\n",
#                 "ALTER TABLE mydeltatablespectestdb.locchange SET LOCATION "
#                 '"dbfs:/tmp/somewhere/locchange/new"',
#             ],
#         )

#     def test_07_from_tc(self):
#         c = Configurator()
#         c.clear_all_configurations()
#         c.add_sql_resource_path(sql)

#         ds = DeltaTableSpec.from_tc("Table1")
#         self.assertEqual(
#             ds.get_sql_create(),
#             (
#                 "CREATE TABLE sometable\n"
#                 "(\n"
#                 "  a int\n"
#                 ")\n"
#                 "USING DELTA\n"
#                 "TBLPROPERTIES (\n"
#                 '  "delta.columnMapping.mode" = "name",\n'
#                 '  "delta.minReaderVersion" = "2",\n'
#                 '  "delta.minWriterVersion" = "5"\n'
#                 ")\n"
#             ),
#         )

#         # The deltaspec needs to be resolved explicitly
#         ds = DeltaTableSpec.from_tc("Table2")
#         c.set_prod()
#         self.assertEqual(ds.fully_substituted().location, "dbfs:/mnt/foo/bar")
#         # This allows it to be globally initialized before the prod/debug is set
#         c.set_debug()
#         self.assertEqual(ds.fully_substituted().location, "dbfs:/tmp/foo/bar")

#     def test_08_modify_to_allow_drop(self):
#         Configurator().set_prod()
#         # a table has been created without the DeltaTableSpec:
#         base = DeltaTableSpec(
#             name="myDeltaTableSpecTestDb{ID}.direct",
#             schema=t.StructType(
#                 fields=[
#                     t.StructField(name="b", dataType=t.StringType()),
#                     t.StructField(name="c", dataType=t.DoubleType()),
#                     t.StructField(name="d", dataType=t.StringType()),
#                 ]
#             ),
#             tblproperties={
#                 "delta.minReaderVersion": "2",
#                 "delta.minWriterVersion": "3",
#             },
#         )

#         # now a modified version is to be applied that drops columns:
#         ds = DeltaTableSpec.from_sql(tables.simple_modified_sql)
#         diff = ds.compare_to(base)
#         self.assertEqual(
#             diff.alter_statements(allow_columns_drop=True),
#             [
#                 "ALTER TABLE mydeltatablespectestdb.direct SET TBLPROPERTIES "
#                 '("delta.minWriterVersion" = "5", '
#                   '"delta.columnMapping.mode" = "name")',
#                 "ALTER TABLE mydeltatablespectestdb.direct DROP COLUMN (d)",
#             ],
#         )

#     def test_09_generated_and_default_expression(self):
#         c = Configurator()
#         c.clear_all_configurations()
#         c.add_sql_resource_path(sql)
#         ds = DeltaTableSpec.from_tc("Table3")
#         self.assertEqual(
#             ds.schema,
#             t.StructType(
#                 [
#                     t.StructField("id", t.LongType(), True),
#                     t.StructField("a", t.IntegerType(), True),
#                     t.StructField("b", t.IntegerType(), True),
#                     t.StructField("area", t.IntegerType(), True),
#                 ]
#             ),
#         )
#         diff = DeltaTableSpecDifference(
#             target=ds, base=dataclasses.replace(ds, tblproperties={})
#         )
#         # The diff is not a match, tableproperties, mismatch
#         self.assertFalse(diff.complete_match(), diff)

#         # ignore the tblproperties and it matches
#         diff = diff.ignore("tblproperties")
#         self.assertTrue(diff.complete_match(), diff)



import unittest

from pyspark.sql import functions as f
from pyspark.sql.dataframe import DataFrame

from spetlr import Configurator
from spetlr.cache import CachedLoader, CachedLoaderParameters
from spetlr.delta import DbHandle, DeltaHandle
from spetlr.spark import Spark


class TestError(Exception):
    pass


class ChildCacher(CachedLoader):
    to_be_written: DataFrame
    written: DataFrame = None
    to_be_deleted: DataFrame
    deleted: DataFrame = None
    too_many_rows_was_called: bool = False
    fail_delete: bool = False

    def write_operation(self, df: DataFrame):
        DeltaHandle.from_tc("CachedTestTarget").append(df)
        return df

    def delete_operation(self, df: DataFrame) -> DataFrame:
        target_name = Configurator().table_name("CachedTestTarget")

        df.createOrReplaceTempView("to_be_deleted")
        Spark.get().sql(
            f"""
            MERGE INTO {target_name} AS t
            USING to_be_deleted AS d
            ON t.a==d.a
            WHEN MATCHED
            THEN DELETE
        """
        )
        if self.fail_delete:
            raise TestError("intended fail point")
        return df


class CachedLoaderProvisionalMarkupTests(unittest.TestCase):
    params: CachedLoaderParameters

    @classmethod
    def setUpClass(cls) -> None:
        tc = Configurator()
        tc.clear_all_configurations()
        tc.set_debug()

        tc.register("TestDb", dict(name="cache_sequence_test{ID}"))
        tc.register(
            "CachedTest",
            dict(name="{TestDb}.cachedloader_cache"),
        )
        tc.register(
            "CachedTestTarget",
            dict(name="{TestDb}.cachedloader_target"),
        )
        DbHandle.from_tc("TestDb").create()
        spark = Spark.get()
        spark.sql(
            """
            CREATE TABLE IF NOT EXISTS {CachedTest_name}
            (
                a STRING,
                rowHash INTEGER,
                loadedTime TIMESTAMP,
                deletedTime TIMESTAMP
            )
        """.format(
                **tc.get_all_details()
            )
        )

        spark.sql(
            """
            CREATE TABLE IF NOT EXISTS {CachedTestTarget_name}
            (
                a STRING,
                payload STRING
            )
        """.format(
                **tc.get_all_details()
            )
        )

        cls.params = CachedLoaderParameters(
            cache_table_name=tc.table_name("CachedTest"),
            key_cols=[
                "a",
            ],
        )

        cls.sut = ChildCacher(cls.params)

    @classmethod
    def tearDownClass(cls) -> None:
        DbHandle("TestDb").drop_cascade()

    def test_full_sequence(self):
        spark = Spark.get()
        dh = DeltaHandle.from_tc("CachedTestTarget")

        # once, a row was loaded:
        initial_data = spark.createDataFrame(
            [
                ("id1", "payload"),
            ],
            dh.read().schema,
        )
        self.sut.save(initial_data)

        # and all was good
        self.assertEqual(dh.read().count(), 1)

        # Then an error occurred and we attempted to save an empty dataframe
        df = initial_data.filter(f.lit(False))
        self.assertEqual(df.count(), 0)
        # luckily an engineer managed to stop the process before it went too far
        self.sut.fail_delete = True
        try:
            self.sut.save(df)
        except TestError:
            pass
        self.sut.fail_delete = False

        # The next day all was back to normal, and we resumed saving the correct data
        self.sut.save(initial_data)

        # now we have the correct data in the target system once more
        self.assertEqual(dh.read().count(), 1)

        # Note: This final assert failed until we introduced the provisional markup
        # step in the cached loader.



from .AzureTags import AzureTags
from .DataframeCreator import DataframeCreator
from .DeleteMismatchedSchemas import DeleteMismatchedSchemas
from .DropOldestDuplicates import DropOldestDuplicates
from .GetMergeStatement import GetMergeStatement
from .MockExtractor import MockExtractor
from .MockLoader import MockLoader
from .SelectAndCastColumns import SelectAndCastColumns

__all__ = [
    DataframeCreator,
    MockLoader,
    GetMergeStatement,
    MockExtractor,
    DropOldestDuplicates,
    SelectAndCastColumns,
    DeleteMismatchedSchemas,
    AzureTags,
]



import unittest

from pyspark.sql.utils import AnalysisException

from spetlr import Configurator
from spetlr.delta import DbHandle, DeltaHandle
from spetlr.spark import Spark
from spetlr.testutils.CleanupTestDatabases import CleanupTestDatabases
from tests.cluster.delta import extras
from tests.cluster.delta.SparkExecutor import SparkSqlExecutor


class CleanTestTablesTests(unittest.TestCase):
    """
    This tests simulates the concurrent execution of two tests:

    Test A:
    Spark SQL databases is created with SPETLR Configurator debug IDs.

    Test B:
    Spark SQL databases is created with ANOTHER SPETLR Configurator debug ID.
    The CleanTestDatabases is executed in the tearDownClass.

    Goal:
    Test A databases should not be affected.

    NB: This test utilizes the Configurator in a non-trivial manner.
        It is not recommended to replicate this test or Configurator setup.
    """

    dh = None
    tc = None
    dbh = None

    @classmethod
    def setUpClass(cls):
        # Register the delivery table for the table configurator
        cls.setup_conf()

        cls.dbh = DbHandle
        cls.dh = DeltaHandle

        # Ensure no table is there
        cls.dbh.from_tc("SparkTestDb").drop_cascade()
        cls.dh.from_tc("SparkTestTable1").drop()
        cls.dbh.from_tc("SparkTestDb2").drop_cascade()
        cls.dh.from_tc("SparkTestTable2").drop()

    @classmethod
    def tearDownClass(cls):
        cls.dbh.from_tc("SparkTestDb").drop_cascade()
        cls.dh.from_tc("SparkTestTable1").drop()
        cls.dbh.from_tc("SparkTestDb2").drop_cascade()
        cls.dh.from_tc("SparkTestTable2").drop()

    def test_01_remove_only_current_UUID_tables(self):
        """
        Test remove only the current UUID databases.

        - Creates databases and saves table names.
        - Verifies that tables are readable.
        - Generates a new UUID and recreates the tables (and databases).
        - Verifies that the new tables are readable.
        - Cleans up the new databases.
        - Verifies that the old tables still exist and the new tables are removed.
        """
        # Create tables and save table names
        SparkSqlExecutor().execute_sql_file("test1")
        SparkSqlExecutor().execute_sql_file("test2_debug")
        old_table_name_1 = self.tc.table_name("SparkTestTable1")
        old_table_name_2 = self.tc.table_name("SparkTestTable2")

        # Should be able to read tables here
        old_df_1 = Spark.get().table(old_table_name_1)
        old_df_2 = Spark.get().table(old_table_name_2)

        self.assertEqual(old_df_1.count(), 0)
        self.assertEqual(old_df_2.count(), 0)

        # Create a new UUID
        self.tc.regenerate_unique_id_and_clear_conf()
        self.setup_conf()

        SparkSqlExecutor().execute_sql_file("test1")
        SparkSqlExecutor().execute_sql_file("test2_debug")
        new_table_name_1 = self.tc.table_name("SparkTestTable1")
        new_table_name_2 = self.tc.table_name("SparkTestTable2")

        # Should be able to read tables here
        new_df_1 = Spark.get().table(new_table_name_1)
        new_df_2 = Spark.get().table(new_table_name_2)

        self.assertEqual(new_df_1.count(), 0)
        self.assertEqual(new_df_2.count(), 0)

        # Cleanup new tables
        CleanupTestDatabases()

        # The old tables should still exist
        old_df_1 = Spark.get().table(old_table_name_1)
        old_df_2 = Spark.get().table(old_table_name_2)

        self.assertEqual(old_df_1.count(), 0)
        self.assertEqual(old_df_2.count(), 0)

        # The new tables should be removed
        with self.assertRaises(AnalysisException):
            Spark.get().table(new_table_name_1)

        with self.assertRaises(AnalysisException):
            Spark.get().table(new_table_name_2)

    @classmethod
    def setup_conf(cls):
        cls.tc = Configurator()
        cls.tc.add_resource_path(extras)
        cls.tc.set_debug()



import unittest
from unittest.mock import call, patch

from pyspark.sql.types import IntegerType, StringType, StructField, StructType

from spetlr.deltaspec.DeltaTableSpec import DeltaTableSpec
from spetlr.deltaspec.DeltaTableSpecDifference import DeltaTableSpecDifference
from spetlr.spark import Spark


class TestDeltaTableSpecMakeStorageMatch(unittest.TestCase):
    """
    Test suite for the `make_storage_match` method in DeltaTableSpec.

    This test suite covers the following scenarios:

    1. **test_make_storage_match_no_changes**
       - Ensures that `make_storage_match()`
         does nothing if the table already matches the specification.

    2. **test_make_storage_match_add_column**
       - Verifies that `make_storage_match()` correctly
         executes an `ALTER TABLE ... ADD COLUMN` statement when needed.

    3. **test_make_storage_match_drop_column**
       - Checks that `make_storage_match()` executes an
         `ALTER TABLE ... DROP COLUMN` statement when columns should be removed.

    4. **test_make_storage_match_change_column_type**
       - Ensures that `make_storage_match()` executes an
         `ALTER TABLE ... ALTER COLUMN` statement when column types need modification.

    5. **test_make_storage_match_create_table**
       - Simulates a missing table and verifies that
         `make_storage_match(allow_table_create=True)`
         executes a `CREATE TABLE` statement.

    6. **test_make_storage_match_multiple_changes**
       - Ensures `make_storage_match()` correctly applies multiple
         changes (adding, dropping, altering columns) in a single run.

    Each test mocks Spark SQL execution to verify that the correct
    SQL statements are generated and executed.
    """

    def setUp(self):
        """Setup a sample DeltaTableSpec object with a basic schema."""
        self.schema = StructType(
            [
                StructField("id", IntegerType(), True),
                StructField("name", StringType(), True),
            ]
        )

        self.spec = DeltaTableSpec(
            name="test_db.test_table",
            schema=self.schema,  # Now includes a schema
            location="abfss://storage/tables/test_table",
            tblproperties={
                "delta.minReaderVersion": "2",
                "delta.minWriterVersion": "5",
            },
        )

    @patch.object(Spark, "get")
    @patch.object(DeltaTableSpec, "compare_to_name")
    def test_make_storage_match_no_changes(self, mock_compare_to_name, mock_spark_get):
        """Test that make_storage_match() does nothing if no changes are needed."""

        mock_compare_to_name.return_value.is_different.return_value = False

        self.spec.make_storage_match()

        # Ensure no SQL commands were executed
        mock_spark_get.return_value.sql.assert_not_called()

    @patch.object(Spark, "get")
    @patch.object(DeltaTableSpec, "compare_to_name")
    def test_make_storage_match_add_column(self, mock_compare_to_name, mock_spark_get):
        """Test that make_storage_match() adds a column when needed."""

        mock_compare_to_name.return_value.is_different.return_value = True
        mock_compare_to_name.return_value.alter_statements.return_value = [
            "ALTER TABLE test_db.test_table ADD COLUMN new_col STRING"
        ]

        self.spec.make_storage_match(allow_columns_add=True)

        # Ensure the correct SQL was executed
        mock_spark_get.return_value.sql.assert_called_once_with(
            "ALTER TABLE test_db.test_table ADD COLUMN new_col STRING"
        )

    @patch.object(Spark, "get")
    @patch.object(DeltaTableSpec, "compare_to_name")
    def test_make_storage_match_drop_column(self, mock_compare_to_name, mock_spark_get):
        """Test that make_storage_match() drops a column when allowed."""

        mock_compare_to_name.return_value.is_different.return_value = True
        mock_compare_to_name.return_value.alter_statements.return_value = [
            "ALTER TABLE test_db.test_table DROP COLUMN old_col"
        ]

        self.spec.make_storage_match(allow_columns_drop=True)

        # Ensure the correct SQL was executed
        mock_spark_get.return_value.sql.assert_called_once_with(
            "ALTER TABLE test_db.test_table DROP COLUMN old_col"
        )

    @patch.object(Spark, "get")
    @patch.object(DeltaTableSpec, "compare_to_name")
    def test_make_storage_match_change_column_type(
        self, mock_compare_to_name, mock_spark_get
    ):
        """Test that make_storage_match() alters column types when needed."""

        mock_compare_to_name.return_value.is_different.return_value = True
        mock_compare_to_name.return_value.alter_statements.return_value = [
            "ALTER TABLE test_db.test_table ALTER COLUMN some_col TYPE BIGINT"
        ]

        self.spec.make_storage_match(allow_columns_type_change=True)

        # Ensure the correct SQL was executed
        mock_spark_get.return_value.sql.assert_called_once_with(
            "ALTER TABLE test_db.test_table ALTER COLUMN some_col TYPE BIGINT"
        )

    @patch.object(Spark, "get")
    @patch.object(DeltaTableSpec, "compare_to_name")
    def test_make_storage_match_create_table(
        self, mock_compare_to_name, mock_spark_get
    ):
        """Test that make_storage_match() creates the table if it does not exist."""

        # Simulate a missing table by making compare_to_name() return None for base
        mock_compare_to_name.return_value = DeltaTableSpecDifference(
            base=None, target=self.spec
        )

        # Expected CREATE TABLE SQL
        create_sql = self.spec.get_sql_create()

        # Call make_storage_match with table creation enabled
        self.spec.make_storage_match(allow_table_create=True)

        # Ensure the CREATE TABLE SQL was executed
        mock_spark_get.return_value.sql.assert_called_once_with(create_sql)

    @patch.object(Spark, "get")
    @patch.object(DeltaTableSpec, "compare_to_name")
    def test_make_storage_match_multiple_changes(
        self, mock_compare_to_name, mock_spark_get
    ):
        """Test that make_storage_match() executes multiple
        SQL alterations when needed."""

        mock_compare_to_name.return_value.is_different.return_value = True
        mock_compare_to_name.return_value.alter_statements.return_value = [
            "ALTER TABLE test_db.test_table ADD COLUMN new_col STRING",
            "ALTER TABLE test_db.test_table DROP COLUMN old_col",
            "ALTER TABLE test_db.test_table ALTER COLUMN some_col TYPE BIGINT",
        ]

        self.spec.make_storage_match(
            allow_columns_add=True,
            allow_columns_drop=True,
            allow_columns_type_change=True,
        )

        # Convert expected calls into `call()` objects to match actual calls
        expected_calls = [
            call("ALTER TABLE test_db.test_table ADD COLUMN new_col STRING"),
            call("ALTER TABLE test_db.test_table DROP COLUMN old_col"),
            call("ALTER TABLE test_db.test_table ALTER COLUMN some_col TYPE BIGINT"),
        ]

        # Ensure all expected SQL commands were executed
        mock_spark_get.return_value.sql.assert_has_calls(expected_calls, any_order=True)



from typing import Union

from pyspark.sql import DataFrame
from pyspark.sql.types import DataType

from spetlr.sql.CommonBaseServer import CommonBaseServer


class CosmosBaseServer(CommonBaseServer):
    # this is a typing.Protocol and will never be instantiated directly.
    def execute_sql(self, sql: str):
        pass

    def read_table_by_name(self, table_name: str, schema: DataType = None) -> DataFrame:
        """get container contents to df"""
        pass

    def read_table(self, table_id: str, schema: DataType = None) -> DataFrame:
        """get container data based on Configurator handle"""
        pass

    def write_table_by_name(
        self, df_source: DataFrame, table_name: str, rows_per_partition: int = None
    ):
        """write df to cosmos container. Only append is supported."""
        pass

    def write_table(
        self, df_source: DataFrame, table_id: str, rows_per_partition: int = None
    ):
        """write df to cosmos container based on Configurator handle.
        Only append is supported."""
        pass

    def delete_item(
        self, table_id: str, id: Union[int, str], pk: Union[int, str] = None
    ):
        """delete individual item from cosmos container
        based on Configurator handle."""
        pass

    def delete_container(self, table_id: str):
        """delete entire cosmos container based on Configurator handle."""
        pass

    def delete_container_by_name(self, table_name: str):
        """delete entire cosmos container by name."""
        pass

    def recreate_container_by_name(self, table_name: str):
        """extract container details from existing,
        delete the container, then create it again with the same details.
        based on Configurator handle."""
        pass



# These tests will be activated when the corresponding cluster tests get fixed


# import unittest

# from spetlr import Configurator
# from spetlr.configurator.sql.parse_sql import parse_sql_code_to_config
# from spetlr.deltaspec.DeltaDatabaseSpec import DeltaDatabaseSpec


# class TestTableSpecConversions(unittest.TestCase):
#     def test_DeltaDatabaseSpec_from_sql(self):
#         config = parse_sql_code_to_config(
#             """
#             -- SPETLR.CONFIGURATOR key: MyDbAlias
#             CREATE DATABASE IF NOT EXISTS my_db_name
#             COMMENT "Really great db"
#             LOCATION "/somewhere/over/the/rainbow";
#             """
#         )

#         k, v = config.popitem()
#         self.assertEqual(k, "MyDbAlias")
#         c = Configurator()
#         c.register(k, v)
#         db = DeltaDatabaseSpec.from_tc(k)
#         self.assertEqual(
#             db,
#             DeltaDatabaseSpec(
#                 name="my_db_name",
#                 location="/somewhere/over/the/rainbow",
#                 comment="Really great db",
#             ),
#         )

#     def test_DeltaDatabaseSpec_to_sql(self):
#         db = DeltaDatabaseSpec(
#             name="my_db_name",
#             location="/somewhere/over/the/rainbow",
#             comment="Really great db",
#         )
#         self.assertEqual(
#             db.get_create_sql(),
#             (
#                 "CREATE SCHEMA IF NOT EXISTS my_db_name\n"
#                 '  COMMENT "Really great db"\n'
#                 '  LOCATION "dbfs:/somewhere/over/the/rainbow"'
#             ),
#         )

#     def test_DeltaDatabaseSpec_repr(self):
#         db = DeltaDatabaseSpec(
#             name="my_db_name",
#             location="/somewhere/over/the/rainbow",
#             comment="Really great db",
#         )
#         as_text = repr(db)

#         self.assertEqual(db, eval(as_text))

#     def test_no_location(self):
#         self.assertEqual(
#             DeltaDatabaseSpec.from_sql(
#                 """
#             CREATE DATABASE spark_catalog.mydb
#             COMMENT "nice data"
#         """
#             ),
#             DeltaDatabaseSpec(name="spark_catalog.mydb", comment="nice data"),
#         )

#     def test_only_name(self):
#         self.assertEqual(
#             DeltaDatabaseSpec.from_sql("CREATE DATABASE mydb"),
#             DeltaDatabaseSpec(name="mydb"),
#         )



from functools import reduce
from typing import List

from pyspark.sql import DataFrame

from spetlr.etl import Transformer
from spetlr.etl.types import dataset_group


class UnionTransformer(Transformer):
    """
    This transformer unions multiple DataFrames

    Attributes:
    ----------
        allowMissingColumns : bool
            When the parameter allowMissingColumns is True,
            the set of column names in this and other DataFrame
            can differ; missing columns will be filled with null

    Methods
    -------
    process(df: DataFrame):
        returns the input DataFrame

    process_many(datasets: dataset_group):
        returns the union of all input DataFrames
    """

    def __init__(
        self,
        allowMissingColumns: bool = False,
        dataset_input_keys: List[str] = None,
        dataset_output_key: str = None,
        consume_inputs: bool = True,
    ):
        super().__init__(
            dataset_input_keys=dataset_input_keys,
            dataset_output_key=dataset_output_key,
            consume_inputs=consume_inputs,
        )
        self.allowMissingColumns = allowMissingColumns

    def process(self, df: DataFrame) -> DataFrame:
        return df

    def process_many(self, datasets: dataset_group) -> DataFrame:
        dfs = list(datasets.values())
        return reduce(
            lambda df1, df2: df1.unionByName(
                df2, allowMissingColumns=self.allowMissingColumns
            ),
            dfs,
        )



from pyspark.sql import DataFrame
from pyspark.sql import functions as f

from spetlr.etl import Extractor
from spetlr.etl.extractors.simple_extractor import Readable


class SchemaExtractor(Extractor):
    """This extractor will extract from any object that has a .read() method,
    the resulting data frame will have zero rows and is intended for schema
    manipulation only."""

    def __init__(self, handle: Readable, dataset_key: str = None):
        super().__init__(dataset_key=dataset_key)
        self.handle = handle

    def read(self) -> DataFrame:
        df = self.handle.read()
        df = df.where(f.lit(False))
        return df



from unittest.mock import MagicMock

from pyspark.sql import DataFrame

from spetlr.etl import Extractor


class MockExtractor(Extractor, MagicMock):
    def __init__(self, *args, dataset_key: str = None, df: DataFrame = None, **kwargs):
        Extractor.__init__(self, dataset_key=dataset_key)
        MagicMock.__init__(self, *args, **kwargs)
        self.df = df

    def read(self) -> DataFrame:
        return self.df



import unittest
from datetime import datetime, timedelta
from unittest.mock import Mock, patch

import pandas as pd
from pandas.testing import assert_frame_equal
from pytz import utc

from spetlr.exceptions import SpetlrException
from spetlr.power_bi.PowerBi import PowerBi
from spetlr.power_bi.PowerBiClient import PowerBiClient
from spetlr.power_bi.PowerBiException import PowerBiException


class TestPowerBi(unittest.TestCase):
    @patch("requests.get")
    def test_verify_workspace_success(self, mock_get):
        # Arrange
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "value": [{"id": "614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0", "name": "Finance"}]
        }
        mock_get.return_value = mock_response

        sut = PowerBi(PowerBiClient(), workspace_name="Finance")
        sut.powerbi_url = "test/"

        # Act
        result = sut._verify_workspace()

        # Assert
        self.assertTrue(result)
        self.assertEqual("614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0", sut.workspace_id)

    @patch("requests.get")
    def test_verify_workspace_failure(self, mock_get):
        # Arrange
        mock_response = Mock()
        mock_response.status_code = 401
        mock_response.text = "error"
        mock_get.return_value = mock_response

        sut = PowerBi(PowerBiClient(), workspace_name="Finance")
        sut.powerbi_url = "test/"

        # Act
        with self.assertRaises(PowerBiException) as context:
            sut._verify_workspace()

        # Assert
        self.assertIn("Failed to fetch workspaces", str(context.exception))

    @patch("requests.get")
    def test_verify_dataset_success(self, mock_get):
        # Arrange
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "value": [
                {"id": "b1f0a07e-e348-402c-a2b2-11f3e31181ce", "name": "Invoicing"}
            ]
        }
        mock_get.return_value = mock_response

        sut = PowerBi(
            PowerBiClient(),
            workspace_name="Finance",
            dataset_name="Invoicing",
        )
        sut.powerbi_url = "test/"

        # Act
        result = sut._verify_dataset()

        # Assert
        self.assertTrue(result)
        self.assertEqual("b1f0a07e-e348-402c-a2b2-11f3e31181ce", sut.dataset_id)

    @patch("requests.get")
    def test_verify_dataset_failure(self, mock_get):
        # Arrange
        mock_response = Mock()
        mock_response.status_code = 404
        mock_response.text = "error"
        mock_get.return_value = mock_response

        sut = PowerBi(
            PowerBiClient(),
            workspace_name="Finance",
            dataset_name="Invoicing",
        )
        sut.powerbi_url = "test/"

        # Act
        with self.assertRaises(PowerBiException) as context:
            sut._verify_dataset()

        # Assert
        self.assertIn("Failed to fetch datasets", str(context.exception))

    @patch("requests.get")
    def test_get_refresh_history_success(self, mock_get):
        # Arrange
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "value": [
                {
                    "id": 1,
                    "refreshType": "OnDemand",
                    "status": "Completed",
                    "startTime": "2024-02-26T10:00:00Z",
                    "endTime": "2024-02-26T10:05:00Z",  # winter time: 1 hour
                    "serviceExceptionJson": None,  # difference from UTC
                    "requestId": "74d25c0b-0473-4dd9-96ff-3ca737b072a7",
                    "refreshAttempts": None,
                },
                {
                    "id": 2,
                    "refreshType": "ViaEnhancedApi",
                    "status": "Completed",
                    "startTime": "2024-03-31T00:00:00Z",  # summer time: change from 1
                    "endTime": "2024-03-31T02:00:00Z",  # hour to 2 hours difference
                    "serviceExceptionJson": None,
                    "requestId": "aec28227-f7af-4c2d-a4e6-fcb01cd570ec",
                    "refreshAttempts": None,
                },
            ]
        }

        def requests_get(url, headers):
            nonlocal mock_response
            if url.endswith("refreshes") and headers == "api_header":
                return mock_response
            raise ValueError("Unknown URL!")

        mock_get.side_effect = requests_get

        sut = PowerBi(
            PowerBiClient(),
            workspace_id="test",
            dataset_id="test",
            local_timezone_name="Europe/Copenhagen",
        )
        sut.api_header = "api_header"
        sut.powerbi_url = "test/"
        sut._get_access_token = lambda: True
        expected = pd.DataFrame(
            {
                "RefreshType": ["OnDemand", "ViaEnhancedApi"],
                "Status": ["Completed", "Completed"],
                "Seconds": [
                    300,
                    7200,
                ],  # despite the daylight saving change it
                # correctly shows 7200 seconds !
                "StartTimeLocal": [
                    datetime(2024, 2, 26, 11, 0),
                    datetime(2024, 3, 31, 1, 0),
                ],
                "EndTimeLocal": [
                    datetime(2024, 2, 26, 11, 5),
                    datetime(2024, 3, 31, 4, 0),
                ],
                "Error": [None, None],
                "RequestId": [
                    "74d25c0b-0473-4dd9-96ff-3ca737b072a7",
                    "aec28227-f7af-4c2d-a4e6-fcb01cd570ec",
                ],
                "RefreshId": [1, 2],
                "RefreshAttempts": None,
            }
        )
        expected["RefreshId"] = expected["RefreshId"].astype("Int64")
        expected["Seconds"] = expected["Seconds"].astype("Int64")

        # Act
        result = sut._combine_dataframes(sut._get_refresh_history)

        # Assert
        self.assertIsNotNone(result)
        assert_frame_equal(expected, result.get_pandas_df())

    @patch("requests.get")
    def test_get_refresh_history_failure(self, mock_get):
        # Arrange
        mock_response = Mock()
        mock_response.status_code = 404  # dataset or workspace not found
        mock_get.return_value = mock_response

        sut = PowerBi(PowerBiClient(), workspace_id="test", dataset_id="test")
        sut.powerbi_url = "test/"
        sut._get_access_token = lambda: True

        # Act
        with self.assertRaises(PowerBiException) as context:
            sut._combine_dataframes(sut._get_refresh_history)

        # Assert
        self.assertIn(
            "The specified dataset or workspace cannot be found", str(context.exception)
        )

    @patch("requests.post")
    def test_get_partition_tables_success(self, mock_post):
        # Arrange
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "results": [
                {
                    "tables": [
                        {
                            "rows": [
                                {
                                    "[ID]": 1,
                                    "[TableID]": 1001,
                                    "[Name]": "Invoices",
                                    "[RefreshedTime]": "2024-02-26T10:00:00Z",
                                    "[ModifiedTime]": "2024-02-26T10:05:00Z",
                                    "[Description]": "contains invoices",
                                    "[ErrorMessage]": None,
                                },
                                {
                                    "[ID]": 2,
                                    "[TableID]": 1002,
                                    "[Name]": "Customers",
                                    "[RefreshedTime]": "2024-02-26T15:00:00Z",
                                    "[ModifiedTime]": "2024-02-26T15:05:00Z",
                                    "[Description]": "contains customers",
                                    "[ErrorMessage]": "Refresh error",
                                },
                            ]
                        }
                    ]
                }
            ]
        }
        mock_post.return_value = mock_response

        sut = PowerBi(
            PowerBiClient(),
            workspace_id="test",
            dataset_id="test",
            local_timezone_name="Europe/Copenhagen",
        )
        sut.powerbi_url = "test/"
        sut._get_access_token = lambda: True
        expected = pd.DataFrame(
            {
                "TableName": ["Invoices", "Customers"],
                "RefreshTimeLocal": [
                    datetime(2024, 2, 26, 11, 0),
                    datetime(2024, 2, 26, 16, 0),
                ],
                "ModifiedTimeLocal": [
                    datetime(2024, 2, 26, 11, 5),
                    datetime(2024, 2, 26, 16, 5),
                ],
                "Description": ["contains invoices", "contains customers"],
                "ErrorMessage": [None, "Refresh error"],
                "Id": [1, 2],
                "TableId": [1001, 1002],
            }
        ).sort_values("TableName")
        expected["Id"] = expected["Id"].astype("Int64")
        expected["TableId"] = expected["TableId"].astype("Int64")

        # Act
        result = sut._combine_dataframes(sut._get_partition_tables)

        # Assert
        self.assertIsNotNone(result)
        assert_frame_equal(
            expected, result.get_pandas_df()[list(expected.columns.values)]
        )

    @patch("requests.post")
    def test_get_partition_tables_failure(self, mock_post):
        # Arrange
        mock_response = Mock()
        mock_response.status_code = 401
        mock_post.return_value = mock_response

        sut = PowerBi(PowerBiClient(), workspace_id="test", dataset_id="test")
        sut.powerbi_url = "test/"
        sut._get_access_token = lambda: True

        # Act
        # Also test if PowerBiException inherits from SpetlrException!
        with self.assertRaises(SpetlrException) as context:
            sut._combine_dataframes(sut._get_partition_tables)

        # Assert
        self.assertIn("Failed to fetch partition info", str(context.exception))

    @patch("requests.get")
    def test_combine_datasets_on_workspace_level_with_success(self, mock_get):
        # Arrange
        get_workspaces = {
            "value": [
                {
                    "id": "614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0",
                    "name": "Finance",
                    "isReadOnly": False,
                },
                {
                    "id": "5da990e9-089e-472c-a7fa-4fc3dd096d01",
                    "name": "CRM",
                    "isReadOnly": True,
                },
            ]
        }
        get_finance_datasets = {
            "value": [
                {
                    "id": "b1f0a07e-e348-402c-a2b2-11f3e31181ce",
                    "name": "Invoicing",
                    "configuredBy": "james@contoso.com",
                    "isRefreshable": True,
                    "isEffectiveIdentityRequired": False,
                    "isEffectiveIdentityRolesRequired": False,
                },
                {
                    "id": "2e848e9a-47a3-4b0e-a22a-af35507ec8c4",
                    "name": "Reimbursement",
                    "configuredBy": "olivia@contoso.com",
                    "isRefreshable": True,
                    "isEffectiveIdentityRequired": False,
                    "isEffectiveIdentityRolesRequired": False,
                },
                {
                    "id": "4de28a6f-f7d4-4186-a529-bf6c65e67b31",
                    "name": "Fees",
                    "configuredBy": "evelyn@contoso.com",
                    "isRefreshable": False,
                    "isEffectiveIdentityRequired": False,
                    "isEffectiveIdentityRolesRequired": False,
                },
                {
                    "id": "c3c7591a-35bf-4f25-b03c-9c5ed6cdae14",
                    "name": "Fraud",
                    "configuredBy": "evelyn@contoso.com",
                    "isRefreshable": True,
                    "isEffectiveIdentityRequired": True,
                    "isEffectiveIdentityRolesRequired": False,
                },
            ]
        }
        get_crm_datasets = {
            "value": [
                {
                    "id": "4869cc38-646c-45b6-89fd-62615beb9853",
                    "name": "Strategy",
                    "configuredBy": "mark@contoso.com",
                    "isRefreshable": True,
                    "isEffectiveIdentityRequired": False,
                    "isEffectiveIdentityRolesRequired": True,
                },
                {
                    "id": "68a2565d-4959-4421-b91c-bafe792796e1",
                    "name": "Strategy",
                    "configuredBy": "amelia@contoso.com",
                    "isRefreshable": False,
                    "isEffectiveIdentityRequired": False,
                    "isEffectiveIdentityRolesRequired": True,
                },
            ]
        }

        def requests_get(url, headers):
            mock_response = Mock()
            mock_response.status_code = 200
            mock_response.json.return_value = None
            if headers == "api_header":
                if url.endswith("groups"):
                    mock_response.json.return_value = get_workspaces
                elif (
                    url.endswith("datasets")
                    and "614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0" in url
                ):
                    mock_response.json.return_value = get_finance_datasets
                elif (
                    url.endswith("datasets")
                    and "5da990e9-089e-472c-a7fa-4fc3dd096d01" in url
                ):
                    mock_response.json.return_value = get_crm_datasets
                if mock_response.json.return_value:
                    return mock_response
            raise ValueError("Unknown URL! " + url)

        mock_get.side_effect = requests_get

        sut = PowerBi(
            PowerBiClient(),
            local_timezone_name="Europe/Copenhagen",
        )
        sut.api_header = "api_header"
        sut.powerbi_url = "test/"
        sut._get_access_token = lambda: True
        expected = pd.DataFrame(
            {
                "WorkspaceName": [
                    "CRM",
                    "CRM",
                    "Finance",
                    "Finance",
                    "Finance",
                    "Finance",
                ],
                "DatasetName": [
                    "Strategy",
                    "Strategy",
                    "Fees",
                    "Fraud",
                    "Invoicing",
                    "Reimbursement",
                ],
                "ConfiguredBy": [
                    "amelia@contoso.com",
                    "mark@contoso.com",
                    "evelyn@contoso.com",
                    "evelyn@contoso.com",
                    "james@contoso.com",
                    "olivia@contoso.com",
                ],
                "IsRefreshable": [False, True, False, True, True, True],
                "AddRowsApiEnabled": [False, False, False, False, False, False],
                "IsEffectiveIdentityRequired": [
                    False,
                    False,
                    False,
                    True,
                    False,
                    False,
                ],
                "IsEffectiveIdentityRolesRequired": [
                    True,
                    True,
                    False,
                    False,
                    False,
                    False,
                ],
                "IsOnPremGatewayRequired": [False, False, False, False, False, False],
                "DatasetId": [
                    "68a2565d-4959-4421-b91c-bafe792796e1",
                    "4869cc38-646c-45b6-89fd-62615beb9853",
                    "4de28a6f-f7d4-4186-a529-bf6c65e67b31",
                    "c3c7591a-35bf-4f25-b03c-9c5ed6cdae14",
                    "b1f0a07e-e348-402c-a2b2-11f3e31181ce",
                    "2e848e9a-47a3-4b0e-a22a-af35507ec8c4",
                ],
                "WorkspaceId": [
                    "5da990e9-089e-472c-a7fa-4fc3dd096d01",
                    "5da990e9-089e-472c-a7fa-4fc3dd096d01",
                    "614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0",
                    "614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0",
                    "614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0",
                    "614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0",
                ],
            }
        )

        # Act
        result = sut._combine_datasets()

        # Assert
        self.assertIsNotNone(result)
        assert_frame_equal(expected, result.get_pandas_df())

    @patch("requests.get")
    def test_combine_dataframes_on_workspace_level_with_success(self, mock_get):
        # Arrange
        get_workspaces = {
            "value": [
                {
                    "id": "614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0",
                    "name": "Finance",
                    "isReadOnly": False,
                },
                {
                    "id": "5da990e9-089e-472c-a7fa-4fc3dd096d01",
                    "name": "CRM",
                    "isReadOnly": True,  # excluded
                },
            ]
        }
        get_finance_datasets = {
            "value": [
                {
                    "id": "b1f0a07e-e348-402c-a2b2-11f3e31181ce",
                    "name": "Invoicing",
                    "configuredBy": "james@contoso.com",
                    "isRefreshable": True,
                    "isEffectiveIdentityRequired": False,
                    "isEffectiveIdentityRolesRequired": False,
                },
                {
                    "id": "2e848e9a-47a3-4b0e-a22a-af35507ec8c4",
                    "name": "Reimbursement",
                    "configuredBy": "olivia@contoso.com",
                    "isRefreshable": True,
                    "isEffectiveIdentityRequired": False,
                    "isEffectiveIdentityRolesRequired": False,
                },
                {
                    "id": "4869cc38-646c-45b6-89fd-62615beb9853",  # excluded HTTP 401
                    "name": "Strategy",
                    "configuredBy": "mark@contoso.com",
                    "isRefreshable": True,
                    "isEffectiveIdentityRequired": False,
                    "isEffectiveIdentityRolesRequired": False,
                },
                {
                    "id": "68a2565d-4959-4421-b91c-bafe792796e1",
                    "name": "Strategy",
                    "configuredBy": "amelia@contoso.com",  # exclude_creators
                    "isRefreshable": True,
                    "isEffectiveIdentityRequired": False,
                    "isEffectiveIdentityRolesRequired": False,
                },
                {
                    "id": "8e4413eb-ba48-4a29-a0c2-fc80272d858e",
                    "name": "Test",
                    "configuredBy": None,  # exclude_creators
                    "isRefreshable": True,
                    "isEffectiveIdentityRequired": False,
                    "isEffectiveIdentityRolesRequired": False,
                },
                {
                    "id": "4de28a6f-f7d4-4186-a529-bf6c65e67b31",
                    "name": "Fees",
                    "configuredBy": "evelyn@contoso.com",
                    "isRefreshable": False,  # excluded
                    "isEffectiveIdentityRequired": False,
                    "isEffectiveIdentityRolesRequired": False,
                },
                {
                    "id": "c3c7591a-35bf-4f25-b03c-9c5ed6cdae14",
                    "name": "Fraud",
                    "configuredBy": "evelyn@contoso.com",
                    "isRefreshable": True,
                    "isEffectiveIdentityRequired": True,  # excluded, only in unit test!
                    "isEffectiveIdentityRolesRequired": False,
                },
            ]
        }
        get_invoicing_refresh_history = {
            "value": [
                {
                    "id": 2,
                    "refreshType": "ViaEnhancedApi",
                    "status": "Completed",
                    "startTime": "2024-03-31T00:00:00Z",  # summer time: change from 1
                    "endTime": "2024-03-31T02:00:00Z",  # hour to 2 hours difference
                    "serviceExceptionJson": None,
                    "requestId": "aec28227-f7af-4c2d-a4e6-fcb01cd570ec",
                    "refreshAttempts": None,
                },
                {
                    "id": 1,
                    "refreshType": "OnDemand",
                    "status": "Completed",
                    "startTime": "2024-02-26T10:00:00Z",
                    "endTime": "2024-02-26T10:05:00Z",  # winter time: 1 hour
                    "serviceExceptionJson": None,  # difference from UTC
                    "requestId": "74d25c0b-0473-4dd9-96ff-3ca737b072a7",
                    "refreshAttempts": None,
                },
            ]
        }
        get_reimbursement_refresh_history = {
            "value": [
                {
                    "id": 2,
                    "refreshType": "ViaEnhancedApi",
                    "status": "Completed",
                    "startTime": "2024-02-24T10:00:00Z",
                    "endTime": "2024-02-24T10:09:00Z",
                    "serviceExceptionJson": None,
                    "requestId": "8d315d90-6105-450b-af2b-8654b962db98",
                    "refreshAttempts": None,
                },
                {
                    "id": 1,
                    "refreshType": "ViaApi",
                    "status": "Completed",
                    "startTime": "2024-02-23T10:00:00Z",
                    "endTime": "2024-02-23T10:10:00Z",
                    "serviceExceptionJson": None,
                    "requestId": "11abf327-6e6b-4cdf-a9a0-f443d83680ec",
                    "refreshAttempts": None,
                },
            ]
        }

        def requests_get(url, headers):
            mock_response = Mock()
            mock_response.status_code = 200
            mock_response.json.return_value = None
            if headers == "api_header":
                if url.endswith("groups"):
                    mock_response.json.return_value = get_workspaces
                elif (
                    url.endswith("datasets")
                    and "614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0" in url
                ):
                    mock_response.json.return_value = get_finance_datasets
                elif (
                    url.endswith("refreshes")
                    and "b1f0a07e-e348-402c-a2b2-11f3e31181ce" in url
                ):
                    mock_response.json.return_value = get_invoicing_refresh_history
                elif (
                    url.endswith("refreshes")
                    and "2e848e9a-47a3-4b0e-a22a-af35507ec8c4" in url
                ):
                    mock_response.json.return_value = get_reimbursement_refresh_history
                elif (
                    url.endswith("refreshes")
                    and "4869cc38-646c-45b6-89fd-62615beb9853" in url
                ):
                    mock_response.status_code = 401
                if mock_response.json.return_value or mock_response.status_code != 200:
                    return mock_response
            raise ValueError("Unknown URL! " + url)

        mock_get.side_effect = requests_get

        sut = PowerBi(
            PowerBiClient(),
            local_timezone_name="Europe/Copenhagen",
            exclude_creators=["Amelia@contoso.com", None],
        )
        sut.api_header = "api_header"
        sut.powerbi_url = "test/"
        sut._get_access_token = lambda: True
        expected = pd.DataFrame(
            {
                "WorkspaceName": ["Finance", "Finance", "Finance", "Finance"],
                "DatasetName": [
                    "Invoicing",
                    "Invoicing",
                    "Reimbursement",
                    "Reimbursement",
                ],
                "RefreshType": [
                    "ViaEnhancedApi",
                    "OnDemand",
                    "ViaEnhancedApi",
                    "ViaApi",
                ],
                "Status": ["Completed", "Completed", "Completed", "Completed"],
                "Seconds": [7200, 300, 540, 600],
                "StartTimeLocal": [
                    datetime(2024, 3, 31, 1, 0),
                    datetime(2024, 2, 26, 11, 0),
                    datetime(2024, 2, 24, 11, 0),
                    datetime(2024, 2, 23, 11, 0),
                ],
                "EndTimeLocal": [
                    datetime(2024, 3, 31, 4, 0),
                    datetime(2024, 2, 26, 11, 5),
                    datetime(2024, 2, 24, 11, 9),
                    datetime(2024, 2, 23, 11, 10),
                ],
                "Error": [None, None, None, None],
                "RequestId": [
                    "aec28227-f7af-4c2d-a4e6-fcb01cd570ec",
                    "74d25c0b-0473-4dd9-96ff-3ca737b072a7",
                    "8d315d90-6105-450b-af2b-8654b962db98",
                    "11abf327-6e6b-4cdf-a9a0-f443d83680ec",
                ],
                "RefreshId": [2, 1, 2, 1],
                "RefreshAttempts": None,
                "WorkspaceId": [
                    "614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0",
                    "614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0",
                    "614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0",
                    "614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0",
                ],
                "DatasetId": [
                    "b1f0a07e-e348-402c-a2b2-11f3e31181ce",
                    "b1f0a07e-e348-402c-a2b2-11f3e31181ce",
                    "2e848e9a-47a3-4b0e-a22a-af35507ec8c4",
                    "2e848e9a-47a3-4b0e-a22a-af35507ec8c4",
                ],
            }
        )
        expected["RefreshId"] = expected["RefreshId"].astype("Int64")
        expected["Seconds"] = expected["Seconds"].astype("Int64")

        # Act
        result = sut._combine_dataframes(
            sut._get_refresh_history,
            skip_read_only=True,
            skip_not_refreshable=True,
            skip_effective_identity=True,
        )

        # Assert
        self.assertIsNotNone(result)
        assert_frame_equal(expected, result.get_pandas_df())

    @patch("requests.get")
    def test_get_last_refresh_success(self, mock_get):
        # Arrange
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "value": [
                {
                    "requestId": "2",
                    "id": "2",
                    "refreshType": "ViaEnhancedApi",  # skip an enhanced refresh
                    "startTime": "2024-02-27T17:00:00Z",
                    "endTime": None,
                    "status": "Unknown",
                    "serviceExceptionJson": None,
                },
                {
                    "requestId": "1",
                    "id": "1",
                    "refreshType": "ViaApi",
                    "startTime": "2024-02-26T10:00:00Z",
                    "endTime": "2024-02-26T10:05:00Z",
                    "status": "Completed",
                    "serviceExceptionJson": None,
                },
            ]
        }
        mock_get.return_value = mock_response

        sut = PowerBi(PowerBiClient(), workspace_id="test", dataset_id="test")
        sut.powerbi_url = "test/"
        sut._connect = lambda: True

        # Act
        result = sut._get_last_refresh(deep_check=True)

        # Assert
        self.assertTrue(result)
        self.assertEqual("Completed", sut.last_status)
        self.assertFalse(sut.is_enhanced)
        self.assertIsNone(sut.last_exception)
        self.assertEqual("2024-02-26 10:05:00+00:00", str(sut.last_refresh_utc))
        # average of ViaApi only
        self.assertEqual(5 * 60, sut.last_duration_in_seconds)

    @patch("requests.get")
    def test_get_last_refresh_deep_check_with_success(self, mock_get):
        # Arrange
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "value": [
                {
                    "requestId": "5",
                    "id": "4",
                    "refreshType": "ViaEnhancedApi",  # skip an enhanced refresh
                    "startTime": "2024-02-27T17:00:00Z",
                    "endTime": "2024-02-27T17:05:00Z",
                    "status": "Unknown",
                    "serviceExceptionJson": None,
                },
                {
                    "requestId": "4",
                    "id": "4",
                    "refreshType": "ViaApi",
                    "startTime": "2024-02-26T10:00:00Z",
                    "endTime": "2024-02-26T10:05:00Z",
                    "status": "Completed",
                    "serviceExceptionJson": None,
                },
                {
                    "requestId": "3",
                    "id": "3",
                    "refreshType": "OnDemand",
                    "startTime": "2024-02-25T10:00:00Z",
                    "endTime": "2024-02-25T10:05:00Z",
                    "status": "Completed",
                    "serviceExceptionJson": None,
                },
                {
                    "requestId": "2",
                    "id": "2",
                    "refreshType": "ViaEnhancedApi",
                    "startTime": "2024-02-24T10:00:00Z",
                    "endTime": "2024-02-24T10:09:00Z",
                    "status": "Completed",
                    "serviceExceptionJson": None,
                },
                {
                    "requestId": "1",
                    "id": "1",
                    "refreshType": "ViaApi",
                    "startTime": "2024-02-23T10:00:00Z",
                    "endTime": "2024-02-23T10:10:00Z",
                    "status": "Completed",
                    "serviceExceptionJson": None,
                },
            ]
        }
        mock_get.return_value = mock_response

        sut = PowerBi(PowerBiClient(), workspace_id="test", dataset_id="test")
        sut.powerbi_url = "test/"
        sut._connect = lambda: True
        expected_duration = int(7.5 * 60)  # average duration from all ViaApi

        # Act
        result = sut._get_last_refresh(deep_check=True)

        # Assert
        self.assertTrue(result)
        self.assertEqual("Completed", sut.last_status)
        self.assertFalse(sut.is_enhanced)
        self.assertIsNone(sut.last_exception)
        self.assertEqual("2024-02-26 10:05:00+00:00", str(sut.last_refresh_utc))
        # average of ViaApi only
        self.assertEqual(expected_duration, sut.last_duration_in_seconds)

    @patch("requests.get")
    def test_get_last_refresh_deep_check_normal_with_success(self, mock_get):
        # Arrange
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "value": [
                {
                    "requestId": "5",
                    "id": "4",
                    "refreshType": "ViaEnhancedApi",  # skip an enhanced refresh
                    "startTime": "2024-02-27T17:00:00Z",
                    "endTime": "2024-02-27T17:05:00Z",
                    "status": "Unknown",
                    "serviceExceptionJson": None,
                },
                {
                    "requestId": "4",
                    "id": "4",
                    "refreshType": "ViaEnhancedApi",  # skip an enhanced refresh
                    "startTime": "2024-02-27T16:00:00Z",
                    "endTime": "2024-02-27T16:05:00Z",
                    "status": "Completed",
                    "serviceExceptionJson": None,
                },
                {
                    "requestId": "3",
                    "id": "3",
                    "refreshType": "ViaEnhancedApi",  # skip an enhanced refresh
                    "startTime": "2024-02-27T15:00:00Z",
                    "endTime": "2024-02-27T15:05:00Z",
                    "status": "Completed",
                    "serviceExceptionJson": None,
                },
                {
                    "requestId": "2",
                    "id": "2",
                    "refreshType": "ViaEnhancedApi",  # skip an enhanced refresh
                    "startTime": "2024-02-27T14:00:00Z",
                    "endTime": "2024-02-27T14:09:00Z",
                    "status": "Completed",
                    "serviceExceptionJson": None,
                },
                {
                    "requestId": "1",
                    "id": "1",
                    "refreshType": "ViaEnhancedApi",  # the last is ok
                    "startTime": "2024-02-27T10:00:00Z",
                    "endTime": "2024-02-27T10:11:00Z",
                    "status": "Completed",
                    "serviceExceptionJson": None,
                },
            ]
        }
        mock_get.return_value = mock_response

        sut = PowerBi(PowerBiClient(), workspace_id="test", dataset_id="test")
        sut.powerbi_url = "test/"
        sut._connect = lambda: True

        # Act
        result = sut._get_last_refresh(deep_check=True)
        expected_duration = 0  # average duration from all ViaApi

        # Assert
        self.assertTrue(result)
        self.assertEqual("Completed", sut.last_status)
        self.assertTrue(sut.is_enhanced)
        self.assertIsNone(sut.last_exception)
        self.assertEqual("2024-02-27 10:11:00+00:00", str(sut.last_refresh_utc))
        # average of ViaApi only
        self.assertEqual(expected_duration, sut.last_duration_in_seconds)

    @patch("requests.get")
    @patch("requests.post")
    def test_get_last_refresh_with_tables_success(self, mock_post, mock_get):
        # Arrange
        mock_get_response = Mock()
        mock_get_response.status_code = 200
        mock_get_response.json.return_value = {
            "value": [
                {
                    "requestId": "3",
                    "id": "3",
                    "refreshType": "ViaEnhancedApi",
                    "startTime": "2024-02-27T17:00:00Z",
                    "endTime": "2024-02-27T17:05:00Z",
                    "status": "Unknown",  # skip as an enhanced refresh is in progress
                    "serviceExceptionJson": None,
                },
                {
                    "requestId": "2",
                    "id": "2",
                    "refreshType": "ViaEnhancedApi",
                    "startTime": "2024-02-26T10:00:00Z",
                    "endTime": "2024-02-26T10:05:00Z",
                    "status": "Completed",
                    "serviceExceptionJson": None,
                },
                {
                    "requestId": "1",
                    "id": "1",
                    "refreshType": "OnDemand",
                    "startTime": "2024-02-25T10:00:00Z",
                    "endTime": "2024-02-25T10:05:00Z",
                    "status": "Completed",
                    "serviceExceptionJson": None,
                },
            ]
        }
        mock_post_response = Mock()
        mock_post_response.status_code = 200
        mock_post_response.json.return_value = {
            "results": [
                {
                    "tables": [
                        {
                            "rows": [
                                {
                                    "[ID]": 1,
                                    "[TableID]": 1001,
                                    "[Name]": "Invoices",
                                    "[RefreshedTime]": "2024-02-26T10:50:00+01:00",
                                    "[ModifiedTime]": "2024-02-26T10:55:00+01:00",
                                    "[Description]": "contains invoices",
                                    "[ErrorMessage]": None,
                                },
                                {
                                    "[ID]": 2,
                                    "[TableID]": 1002,
                                    "[Name]": "Customers",
                                    "[RefreshedTime]": "2024-02-26T15:50:00+01:00",
                                    "[ModifiedTime]": "2024-02-26T15:55:00+01:00",
                                    "[Description]": "contains customers",
                                    "[ErrorMessage]": None,
                                },
                            ]
                        }
                    ]
                }
            ]
        }
        mock_get.return_value = mock_get_response
        mock_post.return_value = mock_post_response

        sut = PowerBi(
            PowerBiClient(),
            workspace_id="test",
            dataset_id="test",
            table_names=["Invoices", "Customers"],
        )
        sut.powerbi_url = "test/"
        sut._connect = lambda: True

        # Act
        result = sut._get_last_refresh(deep_check=True)

        # Assert
        self.assertTrue(result)
        self.assertEqual("Completed", sut.last_status)
        self.assertTrue(sut.is_enhanced)
        self.assertEqual("Invoices", sut.table_name)
        self.assertIsNone(sut.last_exception)
        self.assertEqual("2024-02-26 09:50:00+00:00", str(sut.last_refresh_utc))
        self.assertEqual(0, sut.last_duration_in_seconds)  # tables were specified!

    @patch("requests.get")
    @patch("requests.post")
    def test_get_last_refresh_with_failed_tables_success(self, mock_post, mock_get):
        # Arrange
        mock_get_response = Mock()
        mock_get_response.status_code = 200
        mock_get_response.json.return_value = {
            "value": [
                {
                    "requestId": "2",
                    "id": "2",
                    "refreshType": "ViaApi",
                    "startTime": "2024-02-26T10:00:00Z",
                    "endTime": "2024-02-26T10:05:00Z",
                    "status": "Completed",
                    "serviceExceptionJson": None,
                },
                {
                    "requestId": "1",
                    "id": "1",
                    "refreshType": "OnDemand",
                    "startTime": "2024-02-25T10:00:00Z",
                    "endTime": "2024-02-25T10:05:00Z",
                    "status": "Completed",
                    "serviceExceptionJson": None,
                },
            ]
        }
        mock_post_response = Mock()
        mock_post_response.status_code = 200
        mock_post_response.json.return_value = {
            "results": [
                {
                    "tables": [
                        {
                            "rows": [
                                {
                                    "[ID]": 1,
                                    "[TableID]": 1001,
                                    "[Name]": "Invoices",
                                    "[RefreshedTime]": "2024-02-26T10:50:00+01:00",
                                    "[ModifiedTime]": "2024-02-26T10:55:00+01:00",
                                    "[Description]": "contains invoices",
                                    "[ErrorMessage]": None,
                                },
                                {
                                    "[ID]": 2,
                                    "[TableID]": 1002,
                                    "[Name]": "Customers",
                                    "[RefreshedTime]": "2024-02-26T15:50:00+01:00",
                                    "[ModifiedTime]": "2024-02-26T15:55:00+01:00",
                                    "[Description]": "contains customers",
                                    "[ErrorMessage]": "Refresh error",
                                },
                            ]
                        }
                    ]
                }
            ]
        }
        mock_get.return_value = mock_get_response
        mock_post.return_value = mock_post_response

        sut = PowerBi(
            PowerBiClient(),
            workspace_id="test",
            dataset_id="test",
            table_names=["Invoices", "Customers"],
        )
        sut.powerbi_url = "test/"
        sut._connect = lambda: True

        # Act
        result = sut._get_last_refresh()

        # Assert
        self.assertTrue(result)
        self.assertEqual("Failed", sut.last_status)
        self.assertFalse(sut.is_enhanced)
        self.assertEqual("Customers", sut.table_name)
        self.assertEqual("Refresh error", sut.last_exception)
        self.assertIsNone(sut.last_refresh_utc)

    @patch("requests.get")
    @patch("requests.post")
    def test_get_last_refresh_with_unauthorized_tables_success(
        self, mock_post, mock_get
    ):
        # Arrange
        mock_get_response = Mock()
        mock_get_response.status_code = 200
        mock_get_response.json.return_value = {
            "value": [
                {
                    "requestId": "2",
                    "id": "2",
                    "refreshType": "ViaApi",
                    "startTime": "2024-02-26T10:00:00Z",
                    "endTime": "2024-02-26T10:05:00Z",
                    "status": "Completed",
                    "serviceExceptionJson": None,
                },
                {
                    "requestId": "1",
                    "id": "1",
                    "refreshType": "OnDemand",
                    "startTime": "2024-02-25T10:00:00Z",
                    "endTime": "2024-02-25T10:05:00Z",
                    "status": "Completed",
                    "serviceExceptionJson": None,
                },
            ]
        }
        mock_post_response = Mock()
        mock_post_response.status_code = 401
        mock_get.return_value = mock_get_response
        mock_post.return_value = mock_post_response

        sut = PowerBi(
            PowerBiClient(),
            workspace_id="test",
            dataset_id="test",
            table_names=["Invoices", "Customers"],
        )
        sut.powerbi_url = "test/"
        sut._connect = lambda: True

        # Act
        result = sut._get_last_refresh()

        # Assert
        self.assertTrue(result)
        self.assertEqual("Completed", sut.last_status)
        self.assertFalse(sut.is_enhanced)
        self.assertIsNone(sut.table_name)
        self.assertIsNone(sut.last_exception)
        self.assertEqual("2024-02-26 10:05:00+00:00", str(sut.last_refresh_utc))
        self.assertEqual(0, sut.last_duration_in_seconds)  # tables were specified!

    @patch("requests.get")
    def test_get_last_refresh_empty(self, mock_get):
        # Arrange
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {"value": []}
        mock_get.return_value = mock_response

        sut = PowerBi(PowerBiClient(), workspace_id="test", dataset_id="test")
        sut.powerbi_url = "test/"
        sut.last_status = "test"
        sut.last_duration_in_seconds = 5
        sut._connect = lambda: True

        # Act
        result = sut._get_last_refresh()

        # Assert
        self.assertTrue(result)
        self.assertIsNone(sut.last_exception)
        self.assertFalse(sut.is_enhanced)
        self.assertIsNone(sut.last_status)  # must be cleared!
        self.assertEqual(5, sut.last_duration_in_seconds)  # must be kept unchanged!

    @patch("requests.get")
    def test_get_last_refresh_failure(self, mock_get):
        # Arrange
        mock_response = Mock()
        mock_response.status_code = 404  # dataset or workspace not found
        mock_get.return_value = mock_response

        sut = PowerBi(PowerBiClient(), workspace_id="test", dataset_id="test")
        sut.powerbi_url = "test/"
        sut._connect = lambda: True

        # Act
        with self.assertRaises(PowerBiException) as context:
            sut._get_last_refresh()

        # Assert
        self.assertIn(
            "The specified dataset or workspace cannot be found", str(context.exception)
        )

    def test_verify_last_refresh_success(self):
        # Arrange
        sut = PowerBi(
            PowerBiClient(),
            workspace_id="test",
            dataset_id="test",
            max_minutes_after_last_refresh=5,
            local_timezone_name="Europe/Copenhagen",
        )
        sut.last_status = "Completed"
        sut.table_name = "TestTable"
        sut.last_refresh_utc = datetime.now(utc) - timedelta(minutes=1)
        sut.last_refresh_str = str(sut.last_refresh_utc)

        # Act
        result = sut._verify_last_refresh()

        # Assert
        self.assertTrue(result)

    def test_verify_last_refresh_failure(self):
        # Arrange
        sut = PowerBi(
            PowerBiClient(),
            workspace_id="test",
            dataset_id="test",
            max_minutes_after_last_refresh=5,
            local_timezone_name="Europe/Copenhagen",
        )
        sut.last_status = "Completed"
        sut.table_name = "TestTable"
        sut.last_refresh_str = "2024-05-01"
        sut.last_refresh_utc = datetime.now(utc) - timedelta(minutes=10)

        # Act
        with self.assertRaises(PowerBiException) as context:
            sut._verify_last_refresh()

        # Assert
        self.assertIn("finished more than", str(context.exception))
        self.assertIn("TestTable", str(context.exception))

    def test_get_refresh_argument_json_default(self):
        # Arrange
        sut = PowerBi(
            PowerBiClient(),
            workspace_id="614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0",
            dataset_id="b1f0a07e-e348-402c-a2b2-11f3e31181ce",
        )

        # Act
        result = sut._get_refresh_argument_json(True)

        # Assert
        self.assertIsNone(result)

    def test_get_refresh_argument_json_combination_failure(self):
        # Act
        with self.assertRaises(ValueError) as context:
            PowerBi(
                PowerBiClient(),
                workspace_id="614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0",
                dataset_id="b1f0a07e-e348-402c-a2b2-11f3e31181ce",
                table_names=["Invoices", "Customers"],
                mail_on_failure=True,
            )

        # Assert
        self.assertIn("cannot be combined", str(context.exception))

    def test_get_refresh_argument_json_without_tables_and_without_wait(self):
        # Arrange
        sut = PowerBi(
            PowerBiClient(),
            workspace_id="614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0",
            dataset_id="b1f0a07e-e348-402c-a2b2-11f3e31181ce",
            number_of_retries=2,  # should be ignored without tables
            mail_on_failure=True,
        )
        expected_result = "MailOnFailure"

        # Act
        result = sut._get_refresh_argument_json(with_wait=False)

        # Assert
        self.assertIsNotNone(result)
        self.assertFalse("objects" in result)
        self.assertFalse("retryCount" in result)
        self.assertTrue("notifyOption" in result)
        self.assertEqual(expected_result, result["notifyOption"])

    def test_get_refresh_argument_json_with_tables_and_without_wait(self):
        # Arrange
        sut = PowerBi(
            PowerBiClient(),
            workspace_id="614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0",
            dataset_id="b1f0a07e-e348-402c-a2b2-11f3e31181ce",
            table_names=["Invoices", "Customers"],
            number_of_retries=2,  # should not be ignored
        )
        expected_result1 = 2
        expected_result2 = [{"table": "Invoices"}, {"table": "Customers"}]

        # Act
        result = sut._get_refresh_argument_json(with_wait=False)

        # Assert
        self.assertIsNotNone(result)
        self.assertTrue("objects" in result)
        self.assertTrue("retryCount" in result)
        self.assertEqual(expected_result1, result["retryCount"])
        self.assertEqual(expected_result2, result["objects"])

    def test_get_refresh_argument_json_with_tables_and_with_wait(self):
        # Arrange
        sut = PowerBi(
            PowerBiClient(),
            workspace_id="614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0",
            dataset_id="b1f0a07e-e348-402c-a2b2-11f3e31181ce",
            table_names=["Invoices", "Customers"],
            number_of_retries=2,  # should be ignored if with_wait=True
        )
        expected_result = [{"table": "Invoices"}, {"table": "Customers"}]

        # Act
        result = sut._get_refresh_argument_json(with_wait=True)

        # Assert
        self.assertIsNotNone(result)
        self.assertTrue("objects" in result)
        self.assertFalse("retryCount" in result)
        self.assertEqual(expected_result, result["objects"])

    @patch("requests.post")
    def test_trigger_new_refresh_success(self, mock_post):
        # Arrange
        mock_response = Mock()
        mock_response.status_code = 202
        mock_post.return_value = mock_response

        sut = PowerBi(
            PowerBiClient(),
            workspace_id="614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0",
            dataset_id="b1f0a07e-e348-402c-a2b2-11f3e31181ce",
        )
        sut.last_status = "Completed"
        sut.powerbi_url = "test/"

        # Act
        result = sut._trigger_new_refresh()

        # Assert
        self.assertTrue(result)

    @patch("requests.post")
    def test_trigger_new_refresh_failure(self, mock_post):
        # Arrange
        mock_response = Mock()
        mock_response.status_code = 401
        mock_response.text = "error"
        mock_post.return_value = mock_response

        sut = PowerBi(
            PowerBiClient(),
            workspace_id="614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0",
            dataset_id="b1f0a07e-e348-402c-a2b2-11f3e31181ce",
        )
        sut.last_status = "Completed"
        sut.powerbi_url = "test/"

        # Act
        with self.assertRaises(PowerBiException) as context:
            sut._trigger_new_refresh()

        # Assert
        self.assertIn("Failed to trigger a refresh", str(context.exception))

    def test_get_seconds_to_wait_first(self):
        # Arrange
        sut = PowerBi(PowerBiClient())
        sut.last_duration_in_seconds = 0
        elapsed = 5
        expected_result = 15

        # Act
        result = sut._get_seconds_to_wait(elapsed)

        # Assert
        self.assertEqual(expected_result, result)

    def test_get_seconds_to_wait_next(self):
        # Arrange
        sut = PowerBi(PowerBiClient())
        sut.last_duration_in_seconds = 15 * 60
        elapsed = 5
        expected_result = 5 * 60

        # Act
        result = sut._get_seconds_to_wait(elapsed)

        # Assert
        self.assertEqual(expected_result, result)

    def test_get_seconds_to_wait_exceeding_timeout(self):
        # Arrange
        sut = PowerBi(PowerBiClient(), timeout_in_seconds=90)
        sut.last_duration_in_seconds = 15 * 60
        elapsed = 5
        expected_result = 90 - elapsed

        # Act
        result = sut._get_seconds_to_wait(elapsed)

        # Assert
        self.assertEqual(expected_result, result)

    @patch("requests.post")
    def test_refresh_no_restart_no_tables_success(self, mock_post):
        # Arrange
        mock_response = Mock()
        mock_response.status_code = 501  # not called, so the error is irrelevant!
        mock_response.text = "error"
        mock_post.return_value = mock_response

        sut = PowerBi(
            PowerBiClient(),
            workspace_id="614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0",
            dataset_id="b1f0a07e-e348-402c-a2b2-11f3e31181ce",
            timeout_in_seconds=1,
            number_of_retries=0,
        )
        sut.powerbi_url = "test/"
        counter = 0

        def get_last_refresh():
            nonlocal sut, counter
            sut.table_name = None
            sut.last_refresh_str = None
            sut.is_enhanced = False
            counter += 1
            if counter == 1:
                sut.last_status = "Unknown"
            elif counter == 2:
                sut.last_status = "Completed"
                sut.last_refresh_utc = datetime.now(utc)
            else:
                raise ValueError("Called too many times")
            return True

        sut._get_last_refresh = get_last_refresh

        # Act
        sut.refresh()  # No exception!

        # Assert
        self.assertFalse(mock_post.called)

    @patch("requests.post")
    def test_refresh_restart_no_tables_timeout(self, mock_post):
        # Arrange
        mock_response = Mock()
        mock_response.status_code = 202  # called only when restart triggered!
        mock_post.return_value = mock_response

        sut = PowerBi(
            PowerBiClient(),
            workspace_id="614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0",
            dataset_id="b1f0a07e-e348-402c-a2b2-11f3e31181ce",
            timeout_in_seconds=1,
            number_of_retries=0,
        )
        sut.powerbi_url = "test/"
        counter = 0

        def get_last_refresh():
            nonlocal sut, counter
            sut.table_name = None
            sut.last_refresh_str = None
            sut.is_enhanced = True
            counter += 1
            if counter == 1:
                sut.last_status = "Unknown"
            elif counter == 2:
                sut.last_status = "Completed"  # one restart, then timeout!
                sut.last_refresh_utc = datetime(2024, 5, 14, 9, 17, tzinfo=utc)
            else:
                raise ValueError("Called too many times")
            return True

        sut._get_last_refresh = get_last_refresh

        # Act
        with self.assertRaises(PowerBiException) as context:
            sut.refresh()

        # Assert
        self.assertEqual(1, mock_post.call_count)
        self.assertIn("still in progress", str(context.exception))

    @patch("requests.post")
    def test_refresh_no_restart_with_tables_success(self, mock_post):
        # Arrange
        mock_response = Mock()
        mock_response.status_code = 401  # not called, so the error is irrelevant!
        mock_response.text = "error"
        mock_post.return_value = mock_response

        sut = PowerBi(
            PowerBiClient(),
            workspace_id="614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0",
            dataset_id="b1f0a07e-e348-402c-a2b2-11f3e31181ce",
            table_names=["Invoices", "Customers"],
            timeout_in_seconds=1,
            number_of_retries=0,
        )
        sut.powerbi_url = "test/"
        counter = 0

        def get_last_refresh():
            nonlocal sut, counter
            sut.table_name = None
            sut.last_refresh_str = None
            counter += 1
            if counter == 1:
                sut.last_status = "Unknown"
                sut.is_enhanced = False
            else:
                raise ValueError("Called too many times")
            return True

        sut._get_last_refresh = get_last_refresh

        # Act
        sut.refresh()  # No exception!

        # Assert
        self.assertFalse(mock_post.called)

    @patch("requests.post")
    def test_refresh_restart_with_tables_timeout(self, mock_post):
        # Arrange
        mock_response = Mock()
        mock_response.status_code = 202  # called only whe restart triggered!
        mock_post.return_value = mock_response

        sut = PowerBi(
            PowerBiClient(),
            workspace_id="614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0",
            dataset_id="b1f0a07e-e348-402c-a2b2-11f3e31181ce",
            table_names=["Invoices", "Customers"],
            timeout_in_seconds=1,
            number_of_retries=0,
        )
        sut.powerbi_url = "test/"
        counter = 0

        def get_last_refresh():
            nonlocal sut, counter
            sut.table_name = None
            sut.last_refresh_str = None
            sut.is_enhanced = True
            counter += 1
            if counter == 1:
                sut.last_status = "Unknown"
            elif counter == 2:
                sut.last_status = "Completed"  # one restart, then timeout!
                sut.last_refresh_utc = datetime(2024, 5, 14, 9, 17, tzinfo=utc)
            else:
                raise ValueError("Called too many times")
            return True

        sut._get_last_refresh = get_last_refresh

        # Act
        with self.assertRaises(PowerBiException) as context:
            sut.refresh()

        # Assert
        self.assertEqual(1, mock_post.call_count)
        self.assertIn("still in progress", str(context.exception))

    @patch("requests.post")
    def test_refresh_with_retry_timeout(self, mock_post):
        # Arrange
        mock_response = Mock()
        mock_response.status_code = 202  # called twice, in the beginning and on retry!
        mock_post.return_value = mock_response

        sut = PowerBi(
            PowerBiClient(),
            workspace_id="614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0",
            dataset_id="b1f0a07e-e348-402c-a2b2-11f3e31181ce",
            timeout_in_seconds=1,
            number_of_retries=1,
        )
        sut.powerbi_url = "test/"
        counter = 0

        def get_last_refresh():
            nonlocal sut, counter
            sut.table_name = None
            sut.last_refresh_str = None
            sut.is_enhanced = True
            counter += 1
            if counter == 1:
                sut.last_status = "Completed"  # refresh triggered
                sut.last_refresh_utc = datetime(2024, 5, 14, 9, 17, tzinfo=utc)
            elif counter == 2:
                sut.last_status = "Failed"  # refresh triggered again, but then timeout!
            else:
                raise ValueError("Called too many times")
            return True

        sut._get_last_refresh = get_last_refresh

        # Act
        with self.assertRaises(PowerBiException) as context:
            sut.refresh()

        # Assert
        self.assertEqual(2, mock_post.call_count)
        self.assertIn("still in progress", str(context.exception))

    @patch("requests.post")
    def test_refresh_not_retry_failure(self, mock_post):
        # Arrange
        mock_response = Mock()
        mock_response.status_code = 202  # called only in the beginning!
        mock_post.return_value = mock_response

        sut = PowerBi(
            PowerBiClient(),
            workspace_id="614850c2-3a5c-4d2d-bcaa-d3f20f32a2e0",
            dataset_id="b1f0a07e-e348-402c-a2b2-11f3e31181ce",
            timeout_in_seconds=1,
            number_of_retries=0,
        )
        sut.powerbi_url = "test/"
        counter = 0

        def get_last_refresh():
            nonlocal sut, counter
            sut.table_name = None
            sut.last_refresh_str = None
            sut.is_enhanced = True
            counter += 1
            if counter == 1:
                sut.last_status = "Completed"
                sut.last_refresh_utc = datetime(2024, 5, 14, 9, 17, tzinfo=utc)
            elif counter == 2:
                sut.last_status = "Failed"  # the reason of the exception!
            else:
                raise ValueError("Called too many times")
            return True

        sut._get_last_refresh = get_last_refresh

        # Act
        with self.assertRaises(PowerBiException) as context:
            sut.refresh()

        # Assert
        self.assertEqual(1, mock_post.call_count)
        self.assertIn("Last refresh failed", str(context.exception))



import pyspark.sql.types as T
from spetlrtools.testing import DataframeTestCase

from spetlr.etl.transformers import CountryToAlphaCodeTransformer
from spetlr.spark import Spark


class CountryToAlphaCodeTransformerTest(DataframeTestCase):
    def test_country_to_alpha_code_transformer(self):
        inputSchema = T.StructType(
            [
                T.StructField("countryCol", T.StringType(), True),
            ]
        )

        inputData = [
            ("Denmark",),
            ("Germany",),
        ]

        df_input = Spark.get().createDataFrame(data=inputData, schema=inputSchema)

        expectedData = [("DK",), ("DE",)]

        df_transformed = CountryToAlphaCodeTransformer(col_name="countryCol").process(
            df_input
        )

        self.assertDataframeMatches(df_transformed, None, expectedData)

    def test_country_to_alpha_code_transformer_new_col(self):
        inputSchema = T.StructType(
            [
                T.StructField("countryCol", T.StringType(), True),
            ]
        )

        inputData = [
            ("Denmark",),
            ("Germany",),
        ]

        df_input = Spark.get().createDataFrame(data=inputData, schema=inputSchema)

        expectedData = [
            (
                "Denmark",
                "DK",
            ),
            (
                "Germany",
                "DE",
            ),
        ]

        df_transformed = CountryToAlphaCodeTransformer(
            col_name="countryCol", output_col_name="alphaCodeCol"
        ).process(df_input)

        self.assertDataframeMatches(df_transformed, None, expectedData)

    def test_country_to_alpha_code_transformer_failure(self):
        inputSchema = T.StructType(
            [
                T.StructField("countryCol", T.StringType(), True),
            ]
        )

        inputData = [
            ("NotACountry",),
        ]

        df_input = Spark.get().createDataFrame(data=inputData, schema=inputSchema)

        expectedData = [
            ("",),
        ]

        df_transformed = CountryToAlphaCodeTransformer(col_name="countryCol").process(
            df_input
        )

        self.assertDataframeMatches(df_transformed, None, expectedData)



from typing import List
from uuid import uuid4

from spetlr.etl import EtlBase, Orchestrator, dataset_group
from spetlr.etl.loaders import SimpleLoader
from spetlr.etl.loaders.simple_loader import Appendable
from spetlr.etl.transformers import UnionTransformer

from .log_transformer import LogTransformer


class LogOrchestrator(Orchestrator):
    """
    An extension of the Orchestrator class with built-in support for logging using
    LogTransformers.

    The LogOrchestrator class is a specialized subclass of the Orchestrator designed to
    facilitate ETL (Extract, Transform, Load) processes while providing seamless
    integration with LogTransformers. LogTransformers are specialized transformers that
    measure various metrics on dataframes during the ETL process, enabling the
    collection of valuable insights and measurements. Setting dataset output keys on
    log transformers is not required.

    Arguments:
        handles (List[Appendable]): A list of appendable handles that determine the
            sink location of the output generated by LogTransformers.
        suppress_composition_warning (bool, optional):
            Whether to suppress warnings about potential changes in the composition of
            the ETL process. Defaults to False.

    Methods:
        step(etl: EtlBase) -> LogOrchestrator:
            Adds the etl step to the orchestrator. It has the aliases 'extract_from',
            "transform_with', and 'load_into'. The method behaves the same as the step
            method of the Orchestrator class.
        step_log(log_transformers: LogTransformer) -> LogOrchestrator:
            Adds LogTransformers to the orchestrator. It has the alias 'log_with'. This
            step should be used when wanting to add a logging step.
        etl(inputs: dataset_group) ->dataset_group:
            Executes the ETL process. It has the alias 'execute'.
    """

    def __init__(
        self,
        handles: List[Appendable],
        suppress_composition_warning=False,
    ):
        self.handles = handles
        super().__init__(suppress_composition_warning)
        self.log_transformers_output_keys = []

    def step_log(
        self,
        log_transformer: LogTransformer,
    ) -> "LogOrchestrator":
        self.steps.append(log_transformer)

        # the output key of the LogTransformer is added to a list that is later used to
        # automatically load the data. Data is loaded with the handles provided when
        # instantiating the class.
        transformer_output_key = log_transformer.dataset_output_key

        # this actually ensures that even if no output key is set on the LogTransformer
        # the execute method will still operate correctly
        transformer_output_key = transformer_output_key or str(uuid4())
        self.log_transformers_output_keys.append(transformer_output_key)

        return self

    log_with = step_log

    def step(self, etl: EtlBase) -> "LogOrchestrator":
        return super().step(etl)

    extract_from = step
    transform_with = step
    load_into = step

    def etl(self, inputs: dataset_group = None) -> dataset_group:
        # First check if any log transformers have been added.
        # Note that if no log transformers have been added the LogOrchestrator will
        # behave just like a normal orchestrator.
        if len(self.log_transformers_output_keys) != 0:
            # if only a single log transformer has been added its dataset output key is
            # stored in a variable. This can then be written to the handles provided
            # when the class was instantiated.
            if len(self.log_transformers_output_keys) == 1:
                dataset_key_to_log = self.log_transformers_output_keys

            # if more than one log transformers has been added then these need to be
            # unioned before writing them using the handle(s) provided
            if len(self.log_transformers_output_keys) > 1:
                dataset_key_to_log = "UnionTransformer"
                self.steps.append(
                    UnionTransformer(
                        allowMissingColumns=True,
                        dataset_input_keys=self.log_transformers_output_keys,
                        dataset_output_key=dataset_key_to_log,
                        consume_inputs=True,
                    )
                )

            # now the output from the log transformer(s) can be loaded
            for handle in self.handles:
                self.steps.append(
                    SimpleLoader(
                        handle=handle,
                        mode="append",
                        dataset_input_keys=dataset_key_to_log,
                    )
                )

        # finally execute the orchestrator as usual
        return super().etl(inputs)

    execute = etl



import unittest
from typing import Optional

from pyspark.sql import DataFrame

from spetlr.etl import Extractor, Loader, Orchestrator, Transformer
from spetlr.spark import Spark


class BreakFromEtlTests(unittest.TestCase):
    def test_without_break(self):
        with self.assertRaises(NotImplementedError):
            BreakTestOrchestrator(NoBreakExtractor()).execute()

    def test_with_break(self):
        BreakTestOrchestrator(BreakExtractor()).execute()
        self.assertTrue(True)


class NoBreakExtractor(Extractor):
    def read(self) -> DataFrame:
        return Spark.get().createDataFrame([], schema="Id INTEGER")


class BreakExtractor(Extractor):
    def read(self) -> Optional[DataFrame]:
        return None


class BreakTestOrchestrator(Orchestrator):
    def __init__(self, extractor: Extractor):
        super().__init__()
        self.extract_from(extractor)
        self.transform_with(Transformer())
        self.load_into(Loader())
